from numpy import linspace as _np_linspace
from pandas import DataFrame as _pd_DataFrame
from math import pi
from aabpl.illustrations.illustrate_optimal_grid_spacing import ( create_optimal_grid_spacing_gif, )
from aabpl.utils.distances_to_cell import ( get_always_contained_potentially_overlapped_cells,)

def get_next_relevant_grid_spacing(
        relevantGridSizes:_pd_DataFrame,
        # allGridSizes:_pd_DataFrame,
        # current:float=750,
        upper_bound:float=400,
        lower_bound:float=80,
        precision:float=0.1,
        r:float=750,
        smallest_contain_count:int=0, 
        smallest_overlap_count:int=0
)->tuple:
    """
    
    """
    if len(relevantGridSizes) == 0:
        return upper_bound
    
    it = 0
    current = (upper_bound * 9 + lower_bound*1) / 10
    changeDetected = False
    while upper_bound - lower_bound > precision:
        it += 1
        early_it = int(it<20)+int(it<12)+int(it<6)+int(it<3)
        current = (upper_bound * (5+early_it) + lower_bound*(5-early_it)) / 10
        
        
        contain_count, overlap_count = get_always_contained_potentially_overlapped_cells(grid_spacing=current, r=r, countOnly=True)
        #
        changeDetected = (
            relevantGridSizes.iloc[-1]['contain_count'] != contain_count or
            relevantGridSizes.iloc[-1]['overlap_count'] != overlap_count
            )
        
        if changeDetected:
            lower_bound = current
        elif contain_count==smallest_contain_count and smallest_overlap_count==overlap_count:
            # 
            return lower_bound
        else:
            upper_bound = current
        #
    # the current differs in overlapping / intersecting cells return current
    if changeDetected:
        return current
    # else return lower_bound which is <precison close to current
    return lower_bound
#

def find_relevant_grid_spacings(
    largest:float=400,
    smallest:float=80,
    precision:float=.1,
    r:float=750,
    plot_opt_spacing:dict={},     
    GIF:bool=True,     
)->_pd_DataFrame:
    """
    Function that iteratively decreases from 'largest' grid_spacing to 'smallest' in steps of 'stepsize'
    Whenever decreasing the function
    """
    relevantGridSizes = _pd_DataFrame(
        columns=[
            'grid_spacing', 
            'contain_count', 
            'overlap_count',
            'cell_steps_max', 
            'contain_ids', 
            'overlap_ids'
            ]
        )

    current = largest
    print('Start searching for relevant gridsizes.')
    
    smallest_contain_count, smallest_overlap_count = get_always_contained_potentially_overlapped_cells(
        grid_spacing=smallest, r=r, countOnly=True)

    while current > smallest:
        current = get_next_relevant_grid_spacing(
            relevantGridSizes=relevantGridSizes,
            upper_bound = current,
            lower_bound = smallest,
            precision=precision,
            r=r,
            smallest_contain_count=smallest_contain_count, 
            smallest_overlap_count=smallest_overlap_count
        )

        contain_count, overlap_count, contain_ids, overlap_ids, cell_steps_max = get_always_contained_potentially_overlapped_cells(
            grid_spacing=current, r=r, countOnly=False)

        relevantGridSizes.loc[current]= {
                    'grid_spacing': current,
                    'contain_count':contain_count,
                    'overlap_count':overlap_count,
                    'cell_steps_max':cell_steps_max, 
                    'contain_ids':contain_ids, 
                    'overlap_ids':overlap_ids
                    }      

    #

    relevantGridSizes['share_contain'] = (relevantGridSizes['contain_count']*relevantGridSizes['grid_spacing']**2)/(pi*r**2)
    relevantGridSizes['share_overlap'] = (relevantGridSizes['overlap_count']*relevantGridSizes['grid_spacing']**2)/(pi*r**2)
    
    print('Found '+str(len(relevantGridSizes))+' relevant Gridsizes.')

    return relevantGridSizes
#

def fill_in_all_grid_spacing_steps(
        relevantGridSizes:_pd_DataFrame,
        largest:float=400,
        smallest:float=80,
        stepsize:float=.1,
        r:float=750
    ) -> _pd_DataFrame:
    """
    TODO Currently not used
    Takes relevantGridSizes and fills in all missing steps (subject to stepsizes) and return new extended dataframe
    """
    # dec_places = max_decimal_places([smallest, largest, stepsize])
    # allGridSizesVec = [i for i in arange(smallest, largest+stepsize, stepsize)[::-1]]
    allGridSizesVec = [item for sublist in 
                       [[i0]+list(_np_linspace(i0,i1,int((i0-i1)/stepsize)+1)[1:-1]) for (i0,i1) in zip(
                           relevantGridSizes.index[:-1], relevantGridSizes.index[1:]
                           )]
                       for item in sublist]  + [smallest]
    allGridSizes = relevantGridSizes.iloc[:0].copy()
    relevantGridSizesVec_temp = relevantGridSizes.index #['grid_spacing']

    for i in allGridSizesVec:
        selected_ids = relevantGridSizes[relevantGridSizesVec_temp>=i]
        allGridSizes.loc[i] = selected_ids.iloc[-1]
        if i != selected_ids['grid_spacing'].iloc[-1]:
            allGridSizes.loc[i,'share_contain'] = (allGridSizes.loc[i, 'contain_count']*i**2)/(pi*r**2)
            allGridSizes.loc[i,'share_overlap'] = (allGridSizes.loc[i, 'overlap_count']*i**2)/(pi*r**2)
        allGridSizes.loc[i, 'grid_spacing'] = i

    allGridSizes = allGridSizes.astype(relevantGridSizes.dtypes.to_dict()) # ensure correct data types
    print('Extended to allGridSizes returning DataFrame for '+str(len(allGridSizes))+' Gridsizes.')  

    return allGridSizes      
#

def wrap_optimal_grid_fun(
    largest:float=400,
    smallest:float=80,
    stepsize:float=.5,
    precision:float=.01,
    r:float=750,
    plot_opt_spacing:dict={},     
    GIF:bool=True,
    frames_between_relevant:int=6,
    FuncAnimation_interval:int=500,
    FuncAnimation_repeat:bool=True,
    FuncAnimation_repeat_delay:int=1500,
    FuncAnimation_cache_frame_data:bool=True,
    dpi:int=50,
    PillowWriterFps:int=10,
    output_dir:str='./opt_grid_'     
)->_pd_DataFrame:
    """
    Wrapper
    """

    relevantGridSizes = find_relevant_grid_spacings(
        largest=largest,
        smallest=smallest,
        precision=precision,
        r=r
    )
    
    allGridSizes = fill_in_all_grid_spacing_steps(
        relevantGridSizes=relevantGridSizes,
        largest=largest,
        smallest=smallest,
        stepsize=stepsize,
        r=r,
    )
        
    if GIF:
        create_optimal_grid_spacing_gif(
            relevantGridSizes=relevantGridSizes,
            allGridSizes=allGridSizes,
            largest=largest,
            smallest=smallest,
            r=r,
            frames_between_relevant=frames_between_relevant,
            FuncAnimation_interval=FuncAnimation_interval,
            FuncAnimation_repeat=FuncAnimation_repeat,
            FuncAnimation_repeat_delay=FuncAnimation_repeat_delay,
            FuncAnimation_cache_frame_data=FuncAnimation_cache_frame_data,
            dpi=dpi,
            PillowWriterFps=PillowWriterFps,
            output_dir=output_dir
        ) 
    
    return (relevantGridSizes, allGridSizes)
#

def select_optimal_grid_spacing(
        x_min:float=0.,
        x_max:float=11*100.,
        y_min:float=0.,
        y_max:float=2*100.,
        r:float=750,
        n_points:int=1e6,
        n_queries:int=1,
)->float:
    """
    
    """
    grid_spacing = 1.
    print('Selected optimal grid_spacing of: '+str(grid_spacing)+'.')
    relevantGridSizes, allGridsizes = wrap_optimal_grid_fun(
        largest=350,
        smallest=60,
        stepsize=.2,
        precision=1e-13,
        r=r,
        plot_opt_spacing={},     
        GIF=False,
        frames_between_relevant=20,
        FuncAnimation_interval=100,
        FuncAnimation_repeat=True,
        FuncAnimation_repeat_delay=0,
        FuncAnimation_cache_frame_data=False,
        dpi=100,
        PillowWriterFps=8,
        output_dir='./optimal_grid/plots/opt_grid_'     
        )
    
    relevantGridSizes['comp_time_est'] = relevantGridSizes['grid_spacing']


    return grid_spacing
#






from numpy._core.defchararray import (
    equal as equal,
    not_equal as not_equal,
    greater_equal as greater_equal,
    less_equal as less_equal,
    greater as greater,
    less as less,
    str_len as str_len,
    add as add,
    multiply as multiply,
    mod as mod,
    capitalize as capitalize,
    center as center,
    count as count,
    decode as decode,
    encode as encode,
    endswith as endswith,
    expandtabs as expandtabs,
    find as find,
    index as index,
    isalnum as isalnum,
    isalpha as isalpha,
    isdigit as isdigit,
    islower as islower,
    isspace as isspace,
    istitle as istitle,
    isupper as isupper,
    join as join,
    ljust as ljust,
    lower as lower,
    lstrip as lstrip,
    partition as partition,
    replace as replace,
    rfind as rfind,
    rindex as rindex,
    rjust as rjust,
    rpartition as rpartition,
    rsplit as rsplit,
    rstrip as rstrip,
    split as split,
    splitlines as splitlines,
    startswith as startswith,
    strip as strip,
    swapcase as swapcase,
    title as title,
    translate as translate,
    upper as upper,
    zfill as zfill,
    isnumeric as isnumeric,
    isdecimal as isdecimal,
    array as array,
    asarray as asarray,
    compare_chararrays as compare_chararrays,
    chararray as chararray
)

__all__: list[str]



from collections.abc import Iterable
from typing import (
    Literal as L,
    overload,
    TypeVar,
    Any,
    SupportsIndex,
    SupportsInt,
    NamedTuple,
    Generic,
)

import numpy as np
from numpy import (
    generic,
    floating,
    complexfloating,
    signedinteger,
    unsignedinteger,
    timedelta64,
    object_,
    int32,
    float64,
    complex128,
)

from numpy.linalg import LinAlgError as LinAlgError

from numpy._typing import (
    NDArray,
    ArrayLike,
    _ArrayLikeUnknown,
    _ArrayLikeBool_co,
    _ArrayLikeInt_co,
    _ArrayLikeUInt_co,
    _ArrayLikeFloat_co,
    _ArrayLikeComplex_co,
    _ArrayLikeTD64_co,
    _ArrayLikeObject_co,
    DTypeLike,
)

_T = TypeVar("_T")
_ArrayType = TypeVar("_ArrayType", bound=NDArray[Any])
_SCT = TypeVar("_SCT", bound=generic, covariant=True)
_SCT2 = TypeVar("_SCT2", bound=generic, covariant=True)

_2Tuple = tuple[_T, _T]
_ModeKind = L["reduced", "complete", "r", "raw"]

__all__: list[str]

class EigResult(NamedTuple):
    eigenvalues: NDArray[Any]
    eigenvectors: NDArray[Any]

class EighResult(NamedTuple):
    eigenvalues: NDArray[Any]
    eigenvectors: NDArray[Any]

class QRResult(NamedTuple):
    Q: NDArray[Any]
    R: NDArray[Any]

class SlogdetResult(NamedTuple):
    # TODO: `sign` and `logabsdet` are scalars for input 2D arrays and
    # a `(x.ndim - 2)`` dimensionl arrays otherwise
    sign: Any
    logabsdet: Any

class SVDResult(NamedTuple):
    U: NDArray[Any]
    S: NDArray[Any]
    Vh: NDArray[Any]

@overload
def tensorsolve(
    a: _ArrayLikeInt_co,
    b: _ArrayLikeInt_co,
    axes: None | Iterable[int] =...,
) -> NDArray[float64]: ...
@overload
def tensorsolve(
    a: _ArrayLikeFloat_co,
    b: _ArrayLikeFloat_co,
    axes: None | Iterable[int] =...,
) -> NDArray[floating[Any]]: ...
@overload
def tensorsolve(
    a: _ArrayLikeComplex_co,
    b: _ArrayLikeComplex_co,
    axes: None | Iterable[int] =...,
) -> NDArray[complexfloating[Any, Any]]: ...

@overload
def solve(
    a: _ArrayLikeInt_co,
    b: _ArrayLikeInt_co,
) -> NDArray[float64]: ...
@overload
def solve(
    a: _ArrayLikeFloat_co,
    b: _ArrayLikeFloat_co,
) -> NDArray[floating[Any]]: ...
@overload
def solve(
    a: _ArrayLikeComplex_co,
    b: _ArrayLikeComplex_co,
) -> NDArray[complexfloating[Any, Any]]: ...

@overload
def tensorinv(
    a: _ArrayLikeInt_co,
    ind: int = ...,
) -> NDArray[float64]: ...
@overload
def tensorinv(
    a: _ArrayLikeFloat_co,
    ind: int = ...,
) -> NDArray[floating[Any]]: ...
@overload
def tensorinv(
    a: _ArrayLikeComplex_co,
    ind: int = ...,
) -> NDArray[complexfloating[Any, Any]]: ...

@overload
def inv(a: _ArrayLikeInt_co) -> NDArray[float64]: ...
@overload
def inv(a: _ArrayLikeFloat_co) -> NDArray[floating[Any]]: ...
@overload
def inv(a: _ArrayLikeComplex_co) -> NDArray[complexfloating[Any, Any]]: ...

# TODO: The supported input and output dtypes are dependent on the value of `n`.
# For example: `n < 0` always casts integer types to float64
def matrix_power(
    a: _ArrayLikeComplex_co | _ArrayLikeObject_co,
    n: SupportsIndex,
) -> NDArray[Any]: ...

@overload
def cholesky(a: _ArrayLikeInt_co) -> NDArray[float64]: ...
@overload
def cholesky(a: _ArrayLikeFloat_co) -> NDArray[floating[Any]]: ...
@overload
def cholesky(a: _ArrayLikeComplex_co) -> NDArray[complexfloating[Any, Any]]: ...

@overload
def outer(x1: _ArrayLikeUnknown, x2: _ArrayLikeUnknown) -> NDArray[Any]: ...
@overload
def outer(x1: _ArrayLikeBool_co, x2: _ArrayLikeBool_co) -> NDArray[np.bool]: ...
@overload
def outer(x1: _ArrayLikeUInt_co, x2: _ArrayLikeUInt_co) -> NDArray[unsignedinteger[Any]]: ...
@overload
def outer(x1: _ArrayLikeInt_co, x2: _ArrayLikeInt_co) -> NDArray[signedinteger[Any]]: ...
@overload
def outer(x1: _ArrayLikeFloat_co, x2: _ArrayLikeFloat_co) -> NDArray[floating[Any]]: ...
@overload
def outer(
    x1: _ArrayLikeComplex_co,
    x2: _ArrayLikeComplex_co,
) -> NDArray[complexfloating[Any, Any]]: ...
@overload
def outer(
    x1: _ArrayLikeTD64_co,
    x2: _ArrayLikeTD64_co,
    out: None = ...,
) -> NDArray[timedelta64]: ...
@overload
def outer(x1: _ArrayLikeObject_co, x2: _ArrayLikeObject_co) -> NDArray[object_]: ...
@overload
def outer(
    x1: _ArrayLikeComplex_co | _ArrayLikeTD64_co | _ArrayLikeObject_co,
    x2: _ArrayLikeComplex_co | _ArrayLikeTD64_co | _ArrayLikeObject_co,
) -> _ArrayType: ...

@overload
def qr(a: _ArrayLikeInt_co, mode: _ModeKind = ...) -> QRResult: ...
@overload
def qr(a: _ArrayLikeFloat_co, mode: _ModeKind = ...) -> QRResult: ...
@overload
def qr(a: _ArrayLikeComplex_co, mode: _ModeKind = ...) -> QRResult: ...

@overload
def eigvals(a: _ArrayLikeInt_co) -> NDArray[float64] | NDArray[complex128]: ...
@overload
def eigvals(a: _ArrayLikeFloat_co) -> NDArray[floating[Any]] | NDArray[complexfloating[Any, Any]]: ...
@overload
def eigvals(a: _ArrayLikeComplex_co) -> NDArray[complexfloating[Any, Any]]: ...

@overload
def eigvalsh(a: _ArrayLikeInt_co, UPLO: L["L", "U", "l", "u"] = ...) -> NDArray[float64]: ...
@overload
def eigvalsh(a: _ArrayLikeComplex_co, UPLO: L["L", "U", "l", "u"] = ...) -> NDArray[floating[Any]]: ...

@overload
def eig(a: _ArrayLikeInt_co) -> EigResult: ...
@overload
def eig(a: _ArrayLikeFloat_co) -> EigResult: ...
@overload
def eig(a: _ArrayLikeComplex_co) -> EigResult: ...

@overload
def eigh(
    a: _ArrayLikeInt_co,
    UPLO: L["L", "U", "l", "u"] = ...,
) -> EighResult: ...
@overload
def eigh(
    a: _ArrayLikeFloat_co,
    UPLO: L["L", "U", "l", "u"] = ...,
) -> EighResult: ...
@overload
def eigh(
    a: _ArrayLikeComplex_co,
    UPLO: L["L", "U", "l", "u"] = ...,
) -> EighResult: ...

@overload
def svd(
    a: _ArrayLikeInt_co,
    full_matrices: bool = ...,
    compute_uv: L[True] = ...,
    hermitian: bool = ...,
) -> SVDResult: ...
@overload
def svd(
    a: _ArrayLikeFloat_co,
    full_matrices: bool = ...,
    compute_uv: L[True] = ...,
    hermitian: bool = ...,
) -> SVDResult: ...
@overload
def svd(
    a: _ArrayLikeComplex_co,
    full_matrices: bool = ...,
    compute_uv: L[True] = ...,
    hermitian: bool = ...,
) -> SVDResult: ...
@overload
def svd(
    a: _ArrayLikeInt_co,
    full_matrices: bool = ...,
    compute_uv: L[False] = ...,
    hermitian: bool = ...,
) -> NDArray[float64]: ...
@overload
def svd(
    a: _ArrayLikeComplex_co,
    full_matrices: bool = ...,
    compute_uv: L[False] = ...,
    hermitian: bool = ...,
) -> NDArray[floating[Any]]: ...

def svdvals(
    x: _ArrayLikeInt_co | _ArrayLikeFloat_co | _ArrayLikeComplex_co
) -> NDArray[floating[Any]]: ...

# TODO: Returns a scalar for 2D arrays and
# a `(x.ndim - 2)`` dimensionl array otherwise
def cond(x: _ArrayLikeComplex_co, p: None | float | L["fro", "nuc"] = ...) -> Any: ...

# TODO: Returns `int` for <2D arrays and `intp` otherwise
def matrix_rank(
    A: _ArrayLikeComplex_co,
    tol: None | _ArrayLikeFloat_co = ...,
    hermitian: bool = ...,
    *,
    rtol: None | _ArrayLikeFloat_co = ...,
) -> Any: ...

@overload
def pinv(
    a: _ArrayLikeInt_co,
    rcond: _ArrayLikeFloat_co = ...,
    hermitian: bool = ...,
) -> NDArray[float64]: ...
@overload
def pinv(
    a: _ArrayLikeFloat_co,
    rcond: _ArrayLikeFloat_co = ...,
    hermitian: bool = ...,
) -> NDArray[floating[Any]]: ...
@overload
def pinv(
    a: _ArrayLikeComplex_co,
    rcond: _ArrayLikeFloat_co = ...,
    hermitian: bool = ...,
) -> NDArray[complexfloating[Any, Any]]: ...

# TODO: Returns a 2-tuple of scalars for 2D arrays and
# a 2-tuple of `(a.ndim - 2)`` dimensionl arrays otherwise
def slogdet(a: _ArrayLikeComplex_co) -> SlogdetResult: ...

# TODO: Returns a 2-tuple of scalars for 2D arrays and
# a 2-tuple of `(a.ndim - 2)`` dimensionl arrays otherwise
def det(a: _ArrayLikeComplex_co) -> Any: ...

@overload
def lstsq(a: _ArrayLikeInt_co, b: _ArrayLikeInt_co, rcond: None | float = ...) -> tuple[
    NDArray[float64],
    NDArray[float64],
    int32,
    NDArray[float64],
]: ...
@overload
def lstsq(a: _ArrayLikeFloat_co, b: _ArrayLikeFloat_co, rcond: None | float = ...) -> tuple[
    NDArray[floating[Any]],
    NDArray[floating[Any]],
    int32,
    NDArray[floating[Any]],
]: ...
@overload
def lstsq(a: _ArrayLikeComplex_co, b: _ArrayLikeComplex_co, rcond: None | float = ...) -> tuple[
    NDArray[complexfloating[Any, Any]],
    NDArray[floating[Any]],
    int32,
    NDArray[floating[Any]],
]: ...

@overload
def norm(
    x: ArrayLike,
    ord: None | float | L["fro", "nuc"] = ...,
    axis: None = ...,
    keepdims: bool = ...,
) -> floating[Any]: ...
@overload
def norm(
    x: ArrayLike,
    ord: None | float | L["fro", "nuc"] = ...,
    axis: SupportsInt | SupportsIndex | tuple[int, ...] = ...,
    keepdims: bool = ...,
) -> Any: ...

@overload
def matrix_norm(
    x: ArrayLike,
    ord: None | float | L["fro", "nuc"] = ...,
    keepdims: bool = ...,
) -> floating[Any]: ...
@overload
def matrix_norm(
    x: ArrayLike,
    ord: None | float | L["fro", "nuc"] = ...,
    keepdims: bool = ...,
) -> Any: ...

@overload
def vector_norm(
    x: ArrayLike,
    axis: None = ...,
    ord: None | float = ...,
    keepdims: bool = ...,
) -> floating[Any]: ...
@overload
def vector_norm(
    x: ArrayLike,
    axis: SupportsInt | SupportsIndex | tuple[int, ...] = ...,
    ord: None | float = ...,
    keepdims: bool = ...,
) -> Any: ...

# TODO: Returns a scalar or array
def multi_dot(
    arrays: Iterable[_ArrayLikeComplex_co | _ArrayLikeObject_co | _ArrayLikeTD64_co],
    *,
    out: None | NDArray[Any] = ...,
) -> Any: ...

def diagonal(
    x: ArrayLike,  # >= 2D array
    offset: SupportsIndex = ...,
) -> NDArray[Any]: ...

def trace(
    x: ArrayLike,  # >= 2D array
    offset: SupportsIndex = ...,
    dtype: DTypeLike = ...,
) -> Any: ...

@overload
def cross(
    a: _ArrayLikeUInt_co,
    b: _ArrayLikeUInt_co,
    axis: int = ...,
) -> NDArray[unsignedinteger[Any]]: ...
@overload
def cross(
    a: _ArrayLikeInt_co,
    b: _ArrayLikeInt_co,
    axis: int = ...,
) -> NDArray[signedinteger[Any]]: ...
@overload
def cross(
    a: _ArrayLikeFloat_co,
    b: _ArrayLikeFloat_co,
    axis: int = ...,
) -> NDArray[floating[Any]]: ...
@overload
def cross(
    a: _ArrayLikeComplex_co,
    b: _ArrayLikeComplex_co,
    axis: int = ...,
) -> NDArray[complexfloating[Any, Any]]: ...

@overload
def matmul(
    x1: _ArrayLikeInt_co,
    x2: _ArrayLikeInt_co,
) -> NDArray[signedinteger[Any]]: ...
@overload
def matmul(
    x1: _ArrayLikeUInt_co,
    x2: _ArrayLikeUInt_co,
) -> NDArray[unsignedinteger[Any]]: ...
@overload
def matmul(
    x1: _ArrayLikeFloat_co,
    x2: _ArrayLikeFloat_co,
) -> NDArray[floating[Any]]: ...
@overload
def matmul(
    x1: _ArrayLikeComplex_co,
    x2: _ArrayLikeComplex_co,
) -> NDArray[complexfloating[Any, Any]]: ...



from numpy import array as _np_array, zeros as _np_zeros
from numpy.linalg import norm as _np_linalg_norm
from pandas import (DataFrame as _pd_DataFrame, cut as _pd_cut, concat as _pd_concat) 
from aabpl.utils.general import flatten_list
from aabpl.illustrations.illustrate_point_to_disk import illustrate_point_disk
from aabpl.illustrations.plot_pt_vars import create_plots_for_vars
from aabpl.testing.test_performance import time_func_perf


################ aggreagate_point_data_to_disks_vectorized ######################################################################################
@time_func_perf
def aggreagate_point_data_to_disks_vectorized(
    grid:dict,
    pts_source:_pd_DataFrame,
    pts_target:_pd_DataFrame=None,
    r:float=0.0075,
    columns:list=['employment'],
    y:str='lat',
    x:str='lon',
    row_name:str='id_y',
    col_name:str='id_x',
    cell_region_name:str='cell_region',
    sum_suffix:str='_750m',
    exclude_pt_itself:bool=True,
    plot_radius_sums:dict=None,
    plot_pt_disk:dict=None,
    silent = False,
):
    """
    
    """
    if pts_target is None:
        pts_target = pts_source 
    
    # unpack grid_data 
    grid_id_to_pt_ids = grid.id_to_pt_ids
    grid_id_to_sums = grid.id_to_sums
    pt_id_to_row_col = grid.pt_id_to_row_col
    sparse_grid_ids = set(grid_id_to_pt_ids)
    
    region_id_to_contained_cells = grid.search.region_id_to_contained_cells
    region_id_to_overlapped_cells = grid.search.region_id_to_overlapped_cells
    cells_contained_in_all_disks = grid.search.cells_contained_in_all_disks

    pt_id_to_xy_coords = grid.search.target.pt_id_to_xy_coords
    pt_id_to_vals = grid.search.target.pt_id_to_vals
    n_pts = len(pts_source)

    # initialize columns and/or reset to zero 
    sum_radius_names = [(cname+sum_suffix) for cname in columns]
    pts_source[sum_radius_names] = 0
    
  
    sums_within_disks = _np_zeros((n_pts, len(columns)))
    
    if plot_pt_disk is not None:
        if not 'pt_id' in plot_pt_disk:
            plot_pt_disk['pt_id'] = pts_source.index[int(n_pts//2)]
        
    zero_sums = _np_zeros(len(columns),dtype=int) if len(columns) > 1 else 0
    pts_source['initial_sort'] = range(len(pts_source))
    pts_source.sort_values([row_name, col_name, cell_region_name], inplace=True)
    last_pt_row_col = (-1, -1)
    last_cell_region_id = -1
    counter_new_cell = 0
    counter_new_contain_region = 0
    counter_new_overlap_region = 0

    if len(columns) > 1:
        @time_func_perf
        def sum_contained_all_offset_regions(
                pt_row,
                pt_col,
        ):
            cells_cntd_by_pt_cell = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in cells_contained_in_all_disks])
            # cells_cntd_by_pt_cell = [cell_id for cell_id in (id_y_mult*(cells_contained_in_all_disks[:,0]+(pt_row))+(
            #     cells_contained_in_all_disks[:,1]+pt_col)) if cell_id in sparse_grid_ids] 
            if len(cells_cntd_by_pt_cell)>0:
                return _np_array([grid_id_to_sums[g_id] for g_id in cells_cntd_by_pt_cell]).sum(axis=0) 
            return zero_sums
        #
    else:
        @time_func_perf
        def sum_contained_all_offset_regions(
                pt_row,
                pt_col,
        ):
            cells_cntd_by_pt_cell = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in cells_contained_in_all_disks])
            return sum([grid_id_to_sums[g_id] for g_id in cells_cntd_by_pt_cell]) 
        #
    
    if len(columns) > 1:
        @time_func_perf
        def sum_contained_by_offset_region(
                pt_row,
                pt_col,
                cell_region_id,
        ):
            cells_contained_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_contained_cells[cell_region_id]])
            if len(cells_contained_by_pt_region)>0:
                return _np_array([grid_id_to_sums[g_id] for g_id in cells_contained_by_pt_region]).sum(axis=0) 
            return zero_sums
        #
    else:
        @time_func_perf
        def sum_contained_by_offset_region(
                pt_row,
                pt_col,
                cell_region_id,
        ):
            cells_contained_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_contained_cells[cell_region_id]])
            return sum([grid_id_to_sums[g_id] for g_id in cells_contained_by_pt_region]) 
        #
    
    
    @time_func_perf
    def get_pts_overlapped_by_region(
            pt_row,
            pt_col,
            cell_region_id,
    ):  
        
        cells_overlapped_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_overlapped_cells[cell_region_id]])
        return _np_array(flatten_list([
            grid_id_to_pt_ids[cell_id] for cell_id in cells_overlapped_by_pt_region
        ]))
    #
    
    

    if len(columns) > 1:
        @time_func_perf
        def sum_overlapped_pts_in_radius(
            pts_in_cells_overlapped_by_pt_region,
            a_pt_xycoord
        ):
            if len(pts_in_cells_overlapped_by_pt_region) > 0:
                pts_in_radius = pts_in_cells_overlapped_by_pt_region[(_np_linalg_norm( # create a mask of boolean values indicating if point are within radius 
                    _np_array([pt_id_to_xy_coords[pt_id] for pt_id in pts_in_cells_overlapped_by_pt_region]) -
                    a_pt_xycoord, 
                axis=1) <= r)]
                
                return _np_array([pt_id_to_vals[pt_id] for pt_id in pts_in_radius]).sum(axis=0) if len(pts_in_radius) > 0 else zero_sums
                # else no points in radius thus return vector of _np_zeros
            return zero_sums
    else:
        @time_func_perf
        def sum_overlapped_pts_in_radius(
            pts_in_cells_overlapped_by_pt_region,
            a_pt_xycoord
        ):
            if len(pts_in_cells_overlapped_by_pt_region) > 0:
                pts_in_radius = pts_in_cells_overlapped_by_pt_region[(_np_linalg_norm( # create a mask of boolean values indicating if point are within radius 
                    _np_array([pt_id_to_xy_coords[pt_id] for pt_id in pts_in_cells_overlapped_by_pt_region]) -
                    a_pt_xycoord, 
                axis=1) <= r)]
                
                return sum([pt_id_to_vals[pt_id] for pt_id in pts_in_radius]) if len(pts_in_radius) > 0 else 0
                # else no points in radius thus return vector of _np_zeros
            return 0
    

    @time_func_perf
    def do_nothing():
        pass
    
    for (i, pt_id, a_pt_xycoord, (pt_row,pt_col), contain_region_id, overlap_region_id, cell_region_id) in zip(
        range(n_pts),
        pts_source.index,
        pts_source[[x, y,]].values, 
        pts_source[[row_name, col_name]].values,
        pts_source[cell_region_name].values // grid.search.contain_region_mult,
        pts_source[cell_region_name].values % grid.search.contain_region_mult,
        pts_source[cell_region_name].values,
        
        
        ):
        # (pt_row, pt_col) = pt_id_to_row_col[pt_id]
        
        # as pts are sorted by grid cell update only if grid cell changed
        if not (pt_row, pt_col) == last_pt_row_col:
            counter_new_cell += 1
            # cells_cntd_by_pt_cell = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in cells_contained_in_all_disks])
            # # cells_cntd_by_pt_cell = [cell_id for cell_id in (id_y_mult*(cells_contained_in_all_disks[:,0]+(pt_row))+(
            # #     cells_contained_in_all_disks[:,1]+pt_col)) if cell_id in sparse_grid_ids] 
            # sums_cells_cntd_by_pt_cell = (_np_array([grid_id_to_sums[g_id] for g_id in cells_cntd_by_pt_cell]).sum(axis=0) 
            #                           if len(cells_cntd_by_pt_cell)>0 else zero_sums)
            sums_cells_cntd_by_pt_cell = sum_contained_all_offset_regions(pt_row, pt_col)
        #
            
        if not (pt_row, pt_col) == last_pt_row_col or last_contain_region_id != contain_region_id:
            counter_new_contain_region += 1
            # if cell changed or cell region changed
            # this can be improve to capture only changes of cell region conatain ids 
            # cells_contained_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_contained_cells[cell_region_id]])
            # # cells_contained_by_pt_region = [cell_id for cell_id in (id_y_mult*(region_id_to_contained_cells[cell_region_id][:,0]+(pt_row))+(
            # #     region_id_to_contained_cells[cell_region_id][:,1]+pt_col)) if cell_id in sparse_grid_ids]
                           
            # sums_contained_by_pt_region = (_np_array([grid_id_to_sums[g_id] for g_id in cells_contained_by_pt_region]).sum(axis=0) 
            #                                  if len(cells_contained_by_pt_region)>0 else zero_sums)
            sums_contained_by_pt_region = sum_contained_by_offset_region(pt_row, pt_col, cell_region_id)
        do_nothing()

        if not (pt_row, pt_col) == last_pt_row_col or last_overlap_region_id != overlap_region_id:
            counter_new_overlap_region += 1
            # cells_overlapped_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_overlapped_cells[cell_region_id]])
            # # cells_overlapped_by_pt_region = [cell_id for cell_id in (id_y_mult*(region_id_to_overlapped_cells[cell_region_id][:,0]+(pt_row))+(
            # #     region_id_to_overlapped_cells[cell_region_id][:,1]+pt_col)) if cell_id in sparse_grid_ids]
            # pts_in_cells_overlapped_by_pt_region = _np_array(flatten_list([
            #     grid_id_to_pt_ids[cell_id] for cell_id in cells_overlapped_by_pt_region
            # ]))
            pts_in_cells_overlapped_by_pt_region = get_pts_overlapped_by_region(pt_row, pt_col, cell_region_id)
        #
        
        # if len(pts_in_cells_overlapped_by_pt_region) > 0:
        #     pts_in_radius = pts_in_cells_overlapped_by_pt_region[(_np_linalg_norm( # create a mask of boolean values indicating if point are within radius 
        #         _np_array([pt_id_to_xy_coords[pt_id] for pt_id in pts_in_cells_overlapped_by_pt_region]) -
        #         a_pt_xycoord, 
        #     axis=1) <= r)]
            
        #     overlapping_cells_sums = _np_array([pt_id_to_vals[pt_id] for pt_id in pts_in_radius]).sum(axis=0) if len(pts_in_radius) > 0 else zero_sums
        # else:
        #     # else no points in radius thus return vector of _np_zeros
        #     overlapping_cells_sums = zero_sums
        do_nothing()
        
        overlapping_cells_sums = sum_overlapped_pts_in_radius(pts_in_cells_overlapped_by_pt_region, a_pt_xycoord)

        do_nothing()

        # combine sums from three steps.
        # append result 
        sums_within_disks[i,:] = sums_cells_cntd_by_pt_cell + sums_contained_by_pt_region + overlapping_cells_sums
        
        if plot_pt_disk is not None and pt_id == plot_pt_disk['pt_id']:
            cells_cntd_by_pt_cell = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in cells_contained_in_all_disks])
            cells_contained_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_contained_cells[cell_region_id]])
            pts_in_radius = pts_in_cells_overlapped_by_pt_region[(_np_linalg_norm( # create a mask of boolean values indicating if point are within radius 
                _np_array([pt_id_to_xy_coords[pt_id] for pt_id in pts_in_cells_overlapped_by_pt_region]) -
                a_pt_xycoord, 
            axis=1) <= r)]
            illustrate_point_disk(
                grid=grid,
                pts_source=pts_source,
                pts_target=pts_target,
                r=r,
                columns=columns,
                x=x,
                y=y,
                cells_cntd_by_pt_cell=[(row+pt_row,col+pt_col) for row,col in cells_contained_in_all_disks],
                cells_contained_by_pt_region=[(row+pt_row,col+pt_col) for row,col in region_id_to_contained_cells[cell_region_id]],
                cells_overlapped_by_pt_region=[(row+pt_row,col+pt_col) for row,col in region_id_to_overlapped_cells[cell_region_id]],
                pts_in_cell_contained_by_pt_region=_np_array(flatten_list([
                        grid_id_to_pt_ids[cell_id] for cell_id in cells_cntd_by_pt_cell
                    ]+[
                        grid_id_to_pt_ids[cell_id] for cell_id in cells_contained_by_pt_region
                    ])),
                pts_in_cells_overlapped_by_pt_region=pts_in_cells_overlapped_by_pt_region,
                pts_in_radius=pts_in_radius,
                **plot_pt_disk,
            )
        # #

        # set id as last id for next iteration
        last_pt_row_col = (pt_row, pt_col)
        last_contain_region_id = contain_region_id
        last_overlap_region_id = overlap_region_id
    #
    pts_source[sum_radius_names] = pts_source[sum_radius_names].values + sums_within_disks
            
    if exclude_pt_itself and grid.search.tgt_df_contains_src_df:
        # substract data from point itself unless specified otherwise
        pts_source[sum_radius_names] = pts_source[sum_radius_names].values - pts_source[columns]
    
    # print(
    #     "Share of pts in",
    #     "\n- same cell as previous:", 100-int(counter_new_cell/len(pts_source)*100),"%",
    #     "\n- same cell and containing same surrounding cells:",100 - int(counter_new_contain_region/len(pts_source)*100),"%",
    #     "\n- same cell and overlapping same surrounding cells",100 - int(counter_new_overlap_region/len(pts_source)*100),"%")
    def plot_vars(
        self = grid,
        colnames = _np_array([columns, sum_radius_names]), 
        filename:str='',
        **plot_kwargs:dict,
    ):
        return create_plots_for_vars(
            grid=self,
            colnames=colnames,
            filename=filename,
            plot_kwargs=plot_kwargs,
        )

    grid.plot.vars = plot_vars
    
    if plot_radius_sums is not None:
        print('create plot for radius sums')
        create_plots_for_vars(
            grid=grid,
            colnames=_np_array([columns, sum_radius_names]),
            plot_kwargs=plot_radius_sums,
        )
    #
    pts_source.sort_values(['initial_sort'], inplace=True)
    pts_source.drop('initial_sort', axis=1, inplace=True)

    return pts_source[sum_radius_names]
#





"""
Container class for backward compatibility with NumArray.

The user_array.container class exists for backward compatibility with NumArray
and is not meant to be used in new code. If you need to create an array
container class, we recommend either creating a class that wraps an ndarray
or subclasses ndarray.

"""
from numpy._core import (
    array, asarray, absolute, add, subtract, multiply, divide,
    remainder, power, left_shift, right_shift, bitwise_and, bitwise_or,
    bitwise_xor, invert, less, less_equal, not_equal, equal, greater,
    greater_equal, shape, reshape, arange, sin, sqrt, transpose
)


class container:
    """
    container(data, dtype=None, copy=True)

    Standard container-class for easy multiple-inheritance.

    Methods
    -------
    copy
    tostring
    byteswap
    astype

    """
    def __init__(self, data, dtype=None, copy=True):
        self.array = array(data, dtype, copy=copy)

    def __repr__(self):
        if self.ndim > 0:
            return self.__class__.__name__ + repr(self.array)[len("array"):]
        else:
            return self.__class__.__name__ + "(" + repr(self.array) + ")"

    def __array__(self, t=None):
        if t:
            return self.array.astype(t)
        return self.array

    # Array as sequence
    def __len__(self):
        return len(self.array)

    def __getitem__(self, index):
        return self._rc(self.array[index])

    def __setitem__(self, index, value):
        self.array[index] = asarray(value, self.dtype)

    def __abs__(self):
        return self._rc(absolute(self.array))

    def __neg__(self):
        return self._rc(-self.array)

    def __add__(self, other):
        return self._rc(self.array + asarray(other))

    __radd__ = __add__

    def __iadd__(self, other):
        add(self.array, other, self.array)
        return self

    def __sub__(self, other):
        return self._rc(self.array - asarray(other))

    def __rsub__(self, other):
        return self._rc(asarray(other) - self.array)

    def __isub__(self, other):
        subtract(self.array, other, self.array)
        return self

    def __mul__(self, other):
        return self._rc(multiply(self.array, asarray(other)))

    __rmul__ = __mul__

    def __imul__(self, other):
        multiply(self.array, other, self.array)
        return self

    def __div__(self, other):
        return self._rc(divide(self.array, asarray(other)))

    def __rdiv__(self, other):
        return self._rc(divide(asarray(other), self.array))

    def __idiv__(self, other):
        divide(self.array, other, self.array)
        return self

    def __mod__(self, other):
        return self._rc(remainder(self.array, other))

    def __rmod__(self, other):
        return self._rc(remainder(other, self.array))

    def __imod__(self, other):
        remainder(self.array, other, self.array)
        return self

    def __divmod__(self, other):
        return (self._rc(divide(self.array, other)),
                self._rc(remainder(self.array, other)))

    def __rdivmod__(self, other):
        return (self._rc(divide(other, self.array)),
                self._rc(remainder(other, self.array)))

    def __pow__(self, other):
        return self._rc(power(self.array, asarray(other)))

    def __rpow__(self, other):
        return self._rc(power(asarray(other), self.array))

    def __ipow__(self, other):
        power(self.array, other, self.array)
        return self

    def __lshift__(self, other):
        return self._rc(left_shift(self.array, other))

    def __rshift__(self, other):
        return self._rc(right_shift(self.array, other))

    def __rlshift__(self, other):
        return self._rc(left_shift(other, self.array))

    def __rrshift__(self, other):
        return self._rc(right_shift(other, self.array))

    def __ilshift__(self, other):
        left_shift(self.array, other, self.array)
        return self

    def __irshift__(self, other):
        right_shift(self.array, other, self.array)
        return self

    def __and__(self, other):
        return self._rc(bitwise_and(self.array, other))

    def __rand__(self, other):
        return self._rc(bitwise_and(other, self.array))

    def __iand__(self, other):
        bitwise_and(self.array, other, self.array)
        return self

    def __xor__(self, other):
        return self._rc(bitwise_xor(self.array, other))

    def __rxor__(self, other):
        return self._rc(bitwise_xor(other, self.array))

    def __ixor__(self, other):
        bitwise_xor(self.array, other, self.array)
        return self

    def __or__(self, other):
        return self._rc(bitwise_or(self.array, other))

    def __ror__(self, other):
        return self._rc(bitwise_or(other, self.array))

    def __ior__(self, other):
        bitwise_or(self.array, other, self.array)
        return self

    def __pos__(self):
        return self._rc(self.array)

    def __invert__(self):
        return self._rc(invert(self.array))

    def _scalarfunc(self, func):
        if self.ndim == 0:
            return func(self[0])
        else:
            raise TypeError(
                "only rank-0 arrays can be converted to Python scalars.")

    def __complex__(self):
        return self._scalarfunc(complex)

    def __float__(self):
        return self._scalarfunc(float)

    def __int__(self):
        return self._scalarfunc(int)

    def __hex__(self):
        return self._scalarfunc(hex)

    def __oct__(self):
        return self._scalarfunc(oct)

    def __lt__(self, other):
        return self._rc(less(self.array, other))

    def __le__(self, other):
        return self._rc(less_equal(self.array, other))

    def __eq__(self, other):
        return self._rc(equal(self.array, other))

    def __ne__(self, other):
        return self._rc(not_equal(self.array, other))

    def __gt__(self, other):
        return self._rc(greater(self.array, other))

    def __ge__(self, other):
        return self._rc(greater_equal(self.array, other))

    def copy(self):
        ""
        return self._rc(self.array.copy())

    def tostring(self):
        ""
        return self.array.tostring()

    def tobytes(self):
        ""
        return self.array.tobytes()

    def byteswap(self):
        ""
        return self._rc(self.array.byteswap())

    def astype(self, typecode):
        ""
        return self._rc(self.array.astype(typecode))

    def _rc(self, a):
        if len(shape(a)) == 0:
            return a
        else:
            return self.__class__(a)

    def __array_wrap__(self, *args):
        return self.__class__(args[0])

    def __setattr__(self, attr, value):
        if attr == 'array':
            object.__setattr__(self, attr, value)
            return
        try:
            self.array.__setattr__(attr, value)
        except AttributeError:
            object.__setattr__(self, attr, value)

    # Only called after other approaches fail.
    def __getattr__(self, attr):
        if (attr == 'array'):
            return object.__getattribute__(self, attr)
        return self.array.__getattribute__(attr)


#############################################################
# Test of class container
#############################################################
if __name__ == '__main__':
    temp = reshape(arange(10000), (100, 100))

    ua = container(temp)
    # new object created begin test
    print(dir(ua))
    print(shape(ua), ua.shape)  # I have changed Numeric.py

    ua_small = ua[:3, :5]
    print(ua_small)
    # this did not change ua[0,0], which is not normal behavior
    ua_small[0, 0] = 10
    print(ua_small[0, 0], ua[0, 0])
    print(sin(ua_small) / 3. * 6. + sqrt(ua_small ** 2))
    print(less(ua_small, 103), type(less(ua_small, 103)))
    print(type(ua_small * reshape(arange(15), shape(ua_small))))
    print(reshape(ua_small, (5, 3)))
    print(transpose(ua_small))



from math import ceil as _math_ceil
from matplotlib.pyplot import (subplots as _plt_subplots, figure as _plt_figure)
from matplotlib.patches import Circle as _plt_circle, Rectangle as _plt_Rectangle
from ..utils.general import flatten_list


def plot_cell_pattern(
    contained_cells, 
    overlapped_cells, 
    all_cells,
    r:float,
    grid_spacing:float, 
    add_idxs:bool=True,
    **plot_kwargs,
):

    """
    Illustrate method
    """
    # specify default plot kwargs and add defaults
    plot_kwargs = {
        'fig':None,
        'ax':None,
        's':0.8,
        'color':'#eaa',
        'figsize': (20,30),
        **plot_kwargs
    }
    figsize = plot_kwargs.pop('figsize')
    fig = plot_kwargs.pop('fig')
    ax = plot_kwargs.pop('ax')
    ###### initialize plot  ######################

    if ax is None:
        fig, ax = _plt_subplots(1,1, figsize=figsize)
    ################################################################################################################
    colors = ['#ccc', 'green', 'red']
    for (row, col), color in flatten_list(
        [[((row, col), color) for row, col in cells] for cells,color in zip(
        [all_cells, contained_cells, overlapped_cells],
        colors)]):
        ax.add_patch(_plt_Rectangle(
            xy = (col-0.5, row-0.5), 
            width=1, height=1, 
            linewidth=.7, facecolor=color, edgecolor=color, alpha=0.3
        ))
        if add_idxs and color == colors[0]:
            ax.annotate(text=str(row)+","+str(col), xy=(col,row),horizontalalignment='center')

    ratio = r/grid_spacing
    cell_steps_max = _math_ceil(ratio+1.5)

    ax.set_xlim((-cell_steps_max,+cell_steps_max))
    ax.set_ylim((-cell_steps_max,+cell_steps_max))
    ax.set_aspect('equal', adjustable='box')
    #
#



from numpy import (
    array as _np_array, 
    unique as _np_unique, 
    arange as _np_arange,
    linspace, invert, flip, transpose, 
    concatenate, 
    sign as _np_sign, 
    zeros, min, max, equal, where, 
    logical_or, logical_and, all, newaxis
)
from math import ceil as _math_ceil
# from numpy.random import randint, random
from matplotlib.pyplot import (subplots as _plt_subplots, figure as _plt_figure)
# from matplotlib.animation import FuncAnimation, PillowWriter
from matplotlib.patches import Circle as _plt_circle
from matplotlib.figure import Figure as _plt_Figure
from matplotlib.axes._axes import Axes as _plt_Axes
from .plot_utils import (
    create_grid_cell_patches, 
    create_grid_cell_patches_by_type, 
    create_grid_cell_rectangles,
    create_trgl1_patch, 
    create_buffered_trgl1_patch, 
    create_buffered_square_patch,
    dual_circle_union_patch,
    add_grid_cell_rectangles_by_color,
    add_circle_patches,
    create_debuffered_square_patch,
    create_debuffered_trgl1_patch,
    map_2D_to_rgb
)
from aabpl.utils.distances_to_cell import ( get_cells_relevant_for_disk_by_type, get_cell_farthest_vertex_to_point,  )


# Plot
def illustrate_point_to_cell_region_assignment(
    grid:dict,
    triangle_ids,
    region_ids,
    offset_xy,
    transformed_offset_xy,
    r:float=750,
    include_boundary:bool=False,
    **plot_cell_reg_assign,
):

    """
    Illustrate method
    """
    # specify default plot kwargs and add defaults
    plot_kwargs = {
        'fig':None,
        'ax':None,
        's':0.8,
        'color':'#eaa',
        'figsize': (20,30),
        'example_cell_region':0,
        **plot_cell_reg_assign
    }
    figsize = plot_kwargs.pop('figsize')
    fig = plot_kwargs.pop('fig')
    ax = plot_kwargs.pop('ax')
    example_cell_region = plot_kwargs.pop('example_cell_region')
    example_trgl_region = example_cell_region//100

    # unpack vals
    grid_spacing = grid.spacing
    cell_steps_max = -int(-r/grid_spacing)+1
    trgl_region_ids = _np_array([region_id//100 for region_id in region_ids])
    unique_region_ids = _np_unique([region_id for region_id in grid.search.region_id_to_contained_cells])
    unique_trgl_region_ids = _np_unique([region_id//100 for region_id in unique_region_ids])
    
    contain_region_ids = _np_array([region_id // grid.search.contain_region_mult for region_id in region_ids])
    overlap_region_ids = _np_array([region_id % grid.search.contain_region_mult for region_id in region_ids])
    unique_region_ids = sorted(_np_unique([region_id for region_id in grid.search.region_id_to_contained_cells]))
    
    print('unique region_ids:',len(unique_region_ids),'unique trgl_region_ids:', len(unique_trgl_region_ids))

    (cells_contained_in_all_disks, 
     cells_contained_in_all_trgl_disks, 
     cells_maybe_overlapping_a_disk, 
     cells_maybe_overlapping_a_trgl_disk) = get_cells_relevant_for_disk_by_type(
            grid_spacing=grid_spacing,
            r=r,
            include_boundary=include_boundary,
    )

    ###### initialize plot  ######################

    # fig, axs = _plt_subplots(3,3, figsize=(10,15))
    # flat_axs = axs.flat
    if fig is None:
        fig = _plt_figure(1, figsize=figsize)
    elif type(fig) != _plt_Figure:
        raise TypeError
    flat_axs = []

    flat_axs.append(fig.add_subplot(3,3,len(flat_axs)+1))
    flat_axs.append(fig.add_subplot(3,3,len(flat_axs)+1))
    flat_axs.append(fig.add_subplot(3,3,len(flat_axs)+1))

    flat_axs.append(fig.add_subplot(3,3,len(flat_axs)+1+3))
    flat_axs.append(fig.add_subplot(3,3,len(flat_axs)+1+3))
    flat_axs.append(fig.add_subplot(3,3,len(flat_axs)+1+3))
    
    


    ################################################################################################################
    ax = flat_axs[0]

    add_grid_cell_rectangles_by_color(
        [cells_contained_in_all_disks, cells_maybe_overlapping_a_disk],
        ['green', 'red'],
        ax=ax, grid_spacing=grid_spacing
    )
    
    ax.add_patch(create_buffered_square_patch(side_length=grid_spacing, r=r))
    ax.add_patch(create_debuffered_square_patch(side_length=grid_spacing, r=r, linewidth=2, edgecolor="red", facecolor="None"))

    ################################################################################################################
    ax = flat_axs[1]

    add_grid_cell_rectangles_by_color(
        [cells_maybe_overlapping_a_disk, cells_contained_in_all_disks,cells_contained_in_all_trgl_disks, cells_maybe_overlapping_a_trgl_disk],
        ['#ccc', 'blue', 'green', 'red'],
        ax=ax, grid_spacing=grid_spacing
    )

    ax.add_patch(create_trgl1_patch(side_length=grid_spacing/2, linewidth=2))
    ax.add_patch(create_buffered_trgl1_patch(side_length=grid_spacing/2, r=r, linewidth=2))
    ax.add_patch(create_debuffered_trgl1_patch(side_length=grid_spacing/2, r=r, linewidth=2))

    ################################################################################################################
    ax = flat_axs[2]

    add_grid_cell_rectangles_by_color(
        [
            cells_maybe_overlapping_a_disk, 
            cells_contained_in_all_disks,
            cells_contained_in_all_trgl_disks, 
            grid.search.region_id_to_contained_cells[example_cell_region],
            grid.search.weak_order_tree.cells_contained_in_no_trgl1_disk,
            grid.search.weak_order_tree.cells_overlapped_by_all_trgl1_disks,
            grid.search.region_id_to_overlapped_cells[example_cell_region]
        ],
        [
            '#ccc',
            'pink',
            'blue',
            'green',
            'black',
            'orange',
            'red'
        ],
        ax=ax, grid_spacing=grid_spacing
    )
    filter_mask = _np_arange(len(transformed_offset_xy))[trgl_region_ids==example_trgl_region]
    ax.scatter(x=transformed_offset_xy[filter_mask,0],y=transformed_offset_xy[filter_mask,1],s=fig.get_figheight()/75, color='black')
    
    ################################################################################################################
    # n_cols = len(unique_trgl_region_ids)
    # trgl_combs = sorted(list(set([x%100 for x in unique_region_ids])))
    # n_rows = len(trgl_combs)
        
    # # hundreds = set([x//100 for x in grid.search.region_id_to_contained_cells.keys()])
    # # tens = set([x%100//10 for x in grid.search.region_id_to_contained_cells.keys()])
    # # singles = set([x%10 for x in grid.search.region_id_to_contained_cells.keys()])
 
    # # for i in range(1,8+1):
    # #     tr_ids_i = [x for x in unique_region_ids if i == x%100//10] # tens
    # #     for j in range(1,8+1):
    # #         tr_ids_ij = [x for x in tr_ids_i if j == x%10] # singles
    # for i,trgl_comb in enumerate(trgl_combs):
    #     is_first_col = True
    #     for z in [z for z in sorted(unique_trgl_region_ids)]: # hundreds
    #         current_region_id = int(100*z+trgl_comb)
    #         if current_region_id in unique_region_ids:
    #             flat_axs.append(fig.add_subplot(n_rows*3-10,n_cols,n_cols*(n_rows-5+i)+z+1))
    #             ax = flat_axs[-1]

    #             add_grid_cell_rectangles_by_color(
    #                 [
    #                     cells_maybe_overlapping_a_disk, 
    #                     cells_contained_in_all_disks,
    #                     # cells_contained_in_all_trgl_disks, 
    #                     grid.search.region_id_to_contained_cells[current_region_id],
    #                     # grid.search.weak_order_tree.cells_contained_in_no_trgl1_disk,
    #                     # grid.search.weak_order_tree.cells_overlapped_by_all_trgl1_disks,
    #                     grid.search.region_id_to_overlapped_cells[current_region_id]
    #                 ],
    #                 [
    #                     '#ccc',
    #                     'blue',
    #                     # 'blue',
    #                     'green',
    #                     # 'black',
    #                     # 'orange',
    #                     'red'
    #                 ],
    #                 ax=ax, grid_spacing=grid_spacing
    #             )
    #             if is_first_col:
    #                 ax.set_ylabel(str(trgl_comb))
    #                 is_first_col = False
    #             #
    #         #
    #         if i == 0:
    #             ax.set_title(str(z))
    #         #
    #     #
    # #
    
    unique_contain_region_ids = sorted(_np_unique(contain_region_ids))
    unique_overlap_region_ids = sorted(_np_unique(overlap_region_ids))
    n_cols = _math_ceil(len(unique_region_ids)**.5)
    n_rows = n_cols
    print("n r c", n_rows,n_cols)
    # hundreds = set([x//100 for x in grid.search.region_id_to_contained_cells.keys()])
    # tens = set([x%100//10 for x in grid.search.region_id_to_contained_cells.keys()])
    # singles = set([x%10 for x in grid.search.region_id_to_contained_cells.keys()])
 
    # for i in range(1,8+1):
    #     tr_ids_i = [x for x in unique_region_ids if i == x%100//10] # tens
    #     for j in range(1,8+1):
    #         tr_ids_ij = [x for x in tr_ids_i if j == x%10] # singles
    for i,current_region_id in enumerate(unique_region_ids):
        flat_axs.append(fig.add_subplot(n_rows*3-10,n_cols,n_cols*(n_rows-5)+i+1))
        ax = flat_axs[-1]

        add_grid_cell_rectangles_by_color(
            [
                cells_maybe_overlapping_a_disk, 
                cells_contained_in_all_disks,
                # cells_contained_in_all_trgl_disks, 
                grid.search.region_id_to_contained_cells[current_region_id],
                # grid.search.weak_order_tree.cells_contained_in_no_trgl1_disk,
                # grid.search.weak_order_tree.cells_overlapped_by_all_trgl1_disks,
                grid.search.region_id_to_overlapped_cells[current_region_id]
            ],
            [
                '#ccc',
                'blue',
                # 'blue',
                'green',
                # 'black',
                # 'orange',
                'red'
            ],
            ax=ax, grid_spacing=grid_spacing
        )
        #
    #
    
    ################################################################################################################
    ax = flat_axs[3]
    
    # ax dividie into triangles
    
    
    colorcoded_triangle_1_offset = _np_array([rgb_map_2D(x,y) for (x,y) in transformed_offset_xy/grid_spacing])
    ax.scatter(x=offset_xy[:,1],y=offset_xy[:,0],s=fig.get_figheight()/25, c=colorcoded_triangle_1_offset)
    # ax.scatter(x=offset_xy[:,1],y=offset_xy[:,0],s=fig.get_figheight()/50, c=triangle_ids)
    # triangle add number 
    # pos_0, pos_1 = 0.35*grid_spacing, 0.15*grid_spacing
    # ax.annotate('1', ( pos_0,  pos_1), color='black', weight='bold',fontsize=16, ha='center', va='center')
    # ax.annotate('2', ( pos_1,  pos_0), color='black', weight='bold',fontsize=16, ha='center', va='center')
    # ax.annotate('3', (-pos_1,  pos_0), color='black', weight='bold',fontsize=16, ha='center', va='center')
    # ax.annotate('4', (-pos_0,  pos_1), color='black', weight='bold',fontsize=16, ha='center', va='center')
    # ax.annotate('5', (-pos_0, -pos_1), color='black', weight='bold',fontsize=16, ha='center', va='center')
    # ax.annotate('6', (-pos_1, -pos_0), color='black', weight='bold',fontsize=16, ha='center', va='center')
    # ax.annotate('7', ( pos_1, -pos_0), color='black', weight='bold',fontsize=16, ha='center', va='center')
    # ax.annotate('8', ( pos_0, -pos_1), color='black', weight='bold',fontsize=16, ha='center', va='center')
    
    for trgl_region_id in _np_unique(triangle_ids):
        pos=offset_xy[triangle_ids==trgl_region_id].mean(axis=0)
        ax.annotate(str(trgl_region_id), pos, color='black', weight='bold',fontsize=16, ha='center', va='center')

    ################################################################################################################
    ax = flat_axs[4]
    # ax dividie into regions
    ax.scatter(x=transformed_offset_xy[:,0],y=transformed_offset_xy[:,1],s=fig.get_figheight()/50, c=contain_region_ids, cmap='tab20')
    
    add_circle_patches(
        ax=ax,
        # list_of_cells=[cells_maybe_overlapping_a_trgl_disk,cells_maybe_overlapping_a_trgl_disk],
        list_of_cells=[cells_maybe_overlapping_a_trgl_disk,cells_maybe_overlapping_a_trgl_disk],
        list_of_edgecolors=['red','green'],
        list_of_tuples_check_farthest_closest=[[False,True], [True,False],],
        edgecolor_outside_center_cell=True,
        convex_set_boundaries= _np_array([(0.0,0.0),(0.5,0.0),(0.5,0.5)]),
        grid_spacing=grid_spacing, 
        r=r, 
        linewidth=2
    )

    # for trgl_region_id in _np_unique(contain_region_ids):
    #     pos=transformed_offset_xy[contain_region_ids==trgl_region_id].mean(axis=0)
    #     ax.annotate(str(trgl_region_id), pos, color='black',fontsize=12, ha='center', va='center')

    ax.vlines(x=0, ymin=0, ymax=grid_spacing/2, linewidth=2,color='black')


    ################################################################################################################
    ax = flat_axs[5]

    # ax dividie into regions
    # ax.scatter(x=offset_xy[:,1],y=offset_xy[:,0],s=fig.get_figheight()/50, c=region_ids, cmap='tab20')
    ax.scatter(x=offset_xy[:,1],y=offset_xy[:,0],s=fig.get_figheight()/50, c=contain_region_ids, cmap='tab20')
    # ax.scatter(x=offset_xy[:,1],y=offset_xy[:,0],s=fig.get_figheight()/50, c=overlap_region_ids, cmap='tab20')
    
    add_circle_patches(
        ax=ax,
        list_of_cells=[cells_maybe_overlapping_a_disk,cells_maybe_overlapping_a_disk],
        list_of_edgecolors=['red','green'],
        list_of_tuples_check_farthest_closest=[[False,True], [True,False],],
        edgecolor_outside_center_cell=True,
        convex_set_boundaries= _np_array([(-0.5,-0.5), (0.5,-0.5),(0.5,0.5),(-0.5,0.5)]),
        grid_spacing=grid_spacing, 
        r=r, 
        linewidth=1.5,
        linestyle='-.',
    )
        
    for region_id in _np_unique(region_ids):
        pos=offset_xy[region_ids==region_id].mean(axis=0)
        ax.annotate(str(region_id), pos, color='black',fontsize=8, ha='center', va='center')

    ################################################################################################################
    

    for (i, ax) in enumerate(flat_axs):
        # remove ticks
        if i > 5:
            ax.set_xticks([])
            ax.set_yticks([])
        
        # set limits
        if i < 3:
            ax.set_xlim(-(cell_steps_max+1)*grid_spacing,(cell_steps_max+1)*grid_spacing)
            ax.set_ylim(-(cell_steps_max+1)*grid_spacing,(cell_steps_max+1)*grid_spacing)
        elif i in [3,5]:
            ax.set_xlim(-(.6)*grid_spacing,(.6)*grid_spacing)
            ax.set_ylim(-(.6)*grid_spacing,(.6)*grid_spacing)
        elif i in [4]:
            ax.set_xlim(-(.0)*grid_spacing,(.5)*grid_spacing)
            ax.set_ylim(-(.0)*grid_spacing,(.5)*grid_spacing)
        elif i in [5]:
            ax.set_xlim(-(3.5)*grid_spacing,(3.5)*grid_spacing)
            ax.set_ylim(-(3.5)*grid_spacing,(3.5)*grid_spacing)

        if i in [3,4,5]:
            # add diagionals 
            ax.axline([-.5*grid_spacing,  .0*grid_spacing], [ .5*grid_spacing,  .0*grid_spacing], linewidth=1, color='black')
            ax.axline([ .0*grid_spacing, -.5*grid_spacing], [ .0*grid_spacing,  .5*grid_spacing], linewidth=1, color='black')
            ax.axline([-.5*grid_spacing, -.5*grid_spacing], [ .5*grid_spacing,  .5*grid_spacing], linewidth=1, color='black')
            ax.axline([-.5*grid_spacing,  .5*grid_spacing], [ .5*grid_spacing,  -.5*grid_spacing], linewidth=1, color='black')
        
        # FOR ALL AXS 
        ax.vlines(x=-grid_spacing/2, ymin=-grid_spacing/2, ymax=grid_spacing/2, linewidth=2,color='black')
        ax.vlines(x=grid_spacing/2, ymin=-grid_spacing/2, ymax=grid_spacing/2, linewidth=2,color='black')
        ax.hlines(y=-grid_spacing/2, xmin=-grid_spacing/2, xmax=grid_spacing/2, linewidth=2,color='black')
        ax.hlines(y=grid_spacing/2, xmin=-grid_spacing/2, xmax=grid_spacing/2, linewidth=2,color='black')
        ax.set_aspect('equal', adjustable='box')



#


import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
from pandas import DataFrame as _pd_DataFrame
from numpy import array as _np_array
import math
from pyproj import Transformer
from .random_distribution import get_distribution_for_random_points
from aabpl.testing.test_performance import time_func_perf
from aabpl.radius_search.radius_search_class import DiskSearch
from aabpl.radius_search.grid_class import Grid
from aabpl.illustrations.plot_pt_vars import create_plots_for_vars
from aabpl.illustrations.distribution_plot import create_distribution_plot

def convert_wgs_to_utm(lon: float, lat: float):
    """Based on lat and lng, return best utm epsg-code"""
    # https://gis.stackexchange.com/a/269552
    # convert_wgs_to_utm function, see https://stackoverflow.com/a/40140326/4556479
    # see https://gis.stackexchange.com/a/127432/33092
    utm_band = str((math.floor((lon + 180) / 6 ) % 60) + 1)
    if len(utm_band) == 1:
        utm_band = '0'+utm_band
    if lat >= 0:
        epsg_code = '326' + utm_band
        return epsg_code
    epsg_code = '327' + utm_band
    return epsg_code

def convert_coords_to_local_crs(
        pts,
        x:str='lon',
        y:str='lat',
        proj_x:str='proj_lon',
        proj_y:str='proj_lat',
        initial_crs:str="EPSG:4326",
        target_crs:str='auto',
        silent:bool=False,
) -> str:
    
    
    if target_crs == 'auto': 
        local_crs = 'EPSG:'+str(convert_wgs_to_utm(*pts[[x,y]].mean(axis=0)))
    else:
        local_crs = target_crs
    transformer = Transformer.from_crs(crs_from=initial_crs, crs_to=local_crs, always_xy=True)
    pts[proj_x],pts[proj_y] = transformer.transform(pts[x], pts[y])
    if not silent and initial_crs != local_crs:
        print("Reproject from " +str(initial_crs)+' to '+local_crs)
    return local_crs

def ensure_local_crs(
    pts:_pd_DataFrame=None,
    x:str='lon',
    y:str='lat',
    initial_crs:str='EPSG:4326', 
    target_crs:str='auto',
    silent:bool=True,
):
        
    proj_x = next(('proj_x'+str(i) for i in ['']+list(range(len(pts.columns))) if 'proj_x'+str(i) not in pts.columns))
    proj_y = next(('proj_y'+str(i) for i in ['']+list(range(len(pts.columns))) if 'proj_y'+str(i) not in pts.columns))
    if not target_crs is None:
        local_crs = convert_coords_to_local_crs(pts=pts, initial_crs=initial_crs, target_crs=target_crs, x=x, y=y, proj_x=proj_x, proj_y=proj_y,silent=silent)
        if local_crs == initial_crs:
            pts.drop(columns=[proj_x, proj_y], inplace=True)
        else:
            x = proj_x
            y = proj_y
        return x,y,local_crs
    return x,y,initial_crs

def check_kwargs(
        pts:_pd_DataFrame,
        crs:str,
        r:float,
        columns:list=[],
        x:str='lon',
        y:str='lat',
        row_name:str='id_y',
        col_name:str='id_x',
        pts_target:_pd_DataFrame=None,
        x_tgt:str=None,
        y_tgt:str=None,
        row_name_tgt:str=None,
        col_name_tgt:str=None,
        grid:Grid=None,
        proj_crs:str='auto',
        silent:bool=True,
):
    """
    check shared keyword arguments and apply defaults
    """
    if type(row_name) != str:
        raise TypeError('`row_name` must be of type str. Instead provided of type',type(row_name),row_name)
    if type(col_name) != str:
        raise TypeError('`col_name` must be of type str. Instead provided of type',type(col_name),col_name)
    if row_name_tgt is None:
        row_name_tgt = row_name
    elif type(row_name_tgt) != str:
        raise TypeError('`row_name_tgt` must be of type str. Instead provided of type',type(row_name_tgt),row_name_tgt)
    if col_name_tgt is None:
        col_name_tgt = col_name
    elif type(col_name_tgt) != str:
        raise TypeError('`col_name_tgt` must be of type str. Instead provided of type',type(col_name_tgt),col_name_tgt)
    if type(pts) != _pd_DataFrame:
        raise TypeError('`pts` must be a pandas.DataFrame or None. Instead provided of type',type(pts))
    if type(x) != str:
        raise TypeError('`x` must be of type str. Instead provided of type',type(x),x)
    if type(y) != str:
        raise TypeError('`x` must be of type str. Instead provided of type',type(y),y)
    if not x in pts.columns:
        raise ValueError('`x` (x-coord column name) must be in columns of pts')
    if not y in pts.columns:
        raise ValueError('`y` (y-coord column name) must be in columns of pts')
    if x_tgt is None:
        x_tgt = x
    if y_tgt is None:
        y_tgt = y
    same_target = pts_target is None or pts is pts_target
    if pts_target is None:
        pts_target = pts
    else:
        if type(pts_target) != _pd_DataFrame:
            raise TypeError('`pts_target` must be a pandas.DataFrame or None. Instead provided of type',type(pts_target))
    help_col = None
    if type(columns) == str:
        columns = [columns]
    else:
        if columns is None or len(columns)==0:
            help_col = next(('count'+str(i) for i in (['']+list(range(len(pts_target.columns)))) if not 'helper_col'+str(i) in pts_target.columns))
            pts_target[help_col] = 1
            columns = [help_col]
        try:
            if any([type(column)!=str for column in columns]):
                raise TypeError
        except:
            raise TypeError('`columns` must be either a string of single column name or a list of column name strings')
    if any([not column in pts_target.columns for column in columns]):
        raise ValueError('not all columns(',columns,') are in columns of search target pts_target(',pts.columns,')')
    if not x_tgt in pts_target.columns:
        raise ValueError('`x_tgt` (x-coord column name) must be in columns of pts_target')
    if not y_tgt in pts_target.columns:
        raise ValueError('`y_tgt` (y-coord column name) must be in columns of pts_target')
    
    if proj_crs == 'auto': 
        x_center = (min([pts[x].min(), pts_target[x_tgt].min()])+max([pts[x].max(), pts_target[x_tgt].max()]))/2
        y_center = (min([pts[x].min(), pts_target[x_tgt].min()])+max([pts[x].max(), pts_target[x_tgt].max()]))/2
        local_crs = 'EPSG:'+str(convert_wgs_to_utm(x_center, y_center))
    else:
        local_crs = proj_crs
    if local_crs != proj_crs:
        x,y,local_crs = ensure_local_crs(pts=pts, x=x, y=y, initial_crs=crs, target_crs=proj_crs)
        if not same_target:
            x_tgt,y_tgt,local_crs = ensure_local_crs(pts=pts_target, x=x_tgt, y=y_tgt, initial_crs=crs, target_crs=proj_crs)
        else:
            x_tgt,y_tgt = x,y
    
    # OVERWRITE DEFAULTS
    if grid is None:
        grid = create_auto_grid_for_radius_search(
            pts_source=pts,
            initial_crs=crs,
            local_crs=local_crs,
            r=r,
            x=x,
            y=y,
            pts_target=pts_target,
            x_tgt=x_tgt,
            y_tgt=y_tgt,
            silent=silent,
        )

    return (pts, local_crs, columns, x, y, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, help_col)


# TODO remove cell_region from kwargs
@time_func_perf
def create_auto_grid_for_radius_search(
    pts_source:_pd_DataFrame,
    initial_crs:str,
    local_crs:str,
    r:float,
    x:str='lon',
    y:str='lat',
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    silent:bool=True,
):
    """
    Returns a Grid that covers all points and will 
    - can be used to represent clusters
    - and is leverage for performance gains of radius search 

    Args:
    pts_source (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    x (str):
        column name of x-coordinate (=longtitude) in pts_source (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts_source (default='lat')
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    Returns:
    grid (aabl.Grid):
        a grid covering all points (custom class containing 
    """

    if pts_target is None:
        xmin = pts_source[x].min()
        xmax = pts_source[x].max()
        ymin = pts_source[y].min()
        ymax = pts_source[y].max()
    else:
        if y_tgt is None:
            y_tgt = y
        if x_tgt is None:
            x_tgt = x
        xmin = min([pts_source[x].min(), pts_target[x_tgt].min()])
        xmax = max([pts_source[x].max(), pts_target[x_tgt].max()])
        ymin = min([pts_source[y].min(), pts_target[y_tgt].min()])
        ymax = max([pts_source[y].max(), pts_target[y_tgt].max()])

    return Grid(
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            initial_crs=initial_crs,
            local_crs=local_crs,
            set_fixed_spacing=r/3, # TODO don t set fixed spacing but
            silent=silent,
        )
#

@time_func_perf
def radius_search(
    pts:_pd_DataFrame,
    crs:str,
    r:float,
    columns:list=[],
    exclude_pt_itself:bool=True,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    sum_suffix:str='_r_sum', 
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    row_name_tgt:str=None,
    col_name_tgt:str=None,
    grid:Grid=None,
    proj_crs:str='auto',
    include_boundary:bool=False,
    plot_radius_sums:dict=None,
    plot_pt_disk:dict=None,
    plot_cell_reg_assign:dict=None,
    plot_offset_checks:dict=None,
    plot_offset_regions:dict=None,
    plot_offset_raster:dict=None,
    silent:bool=True,
):
    """
    For all points in DataFrame it searches for all other points (potentially of another DataFrame) within the specified radius and aggregate the values for specified column(s)
    The result will be appended to DataFrame.

    Args:
    pts (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    columns (list or str):
        column(s) in DataFrame for which data within search radius shall be aggregated. If None provided it will simply count the points within the radius. 
    exclude_pt_itself (bool):
        whether the sums within search radius point shall exlclude the point data itself (default=True)
    x (str):
        column name of x-coordinate (=longtitude) in pts (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts (default='lat')
    row_name (str):
        name for column that will be appended to pts indicating grid cell row (default='id_y')
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_y')
    sum_suffix (str):
        suffix used for new column(s) creating by aggregating data of columns , 
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    row_name_tgt (str):
        name for column that will be appended to pts indicating grid cell row. If None its assumed to be same as x (default=None)
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_y')
    proj_crs (crs):
        crs projection into which pts for search (source and target) shall be mapped. If 'auto' local crs will be determined automatically. If None no reprojection will be performed (default='auto')
    grid (aabpl.Grid):
        grid of custom class containing points. If None it will automatically one (default=None)
    include_boundary (bool):
        FEATURE NOT YET IMPLEMENTED. Define whether points that are at the distance of exactly the radius shall be considered within (Default=False)
    plot_radius_sums (dict):
        dictionary with kwargs to create plot for radius sums. If None no plot will be created (default=None)
    plot_pt_disk (dict):
        Only needed for development. Dictionary with kwargs to create plot for example radius search. If None no plot will be created (default=None)
    plot_cell_reg_assign (dict):
        Only needed for development. Dictionary with kwargs to create plot for assginments of points to cell offset regions. If None no plot will be created (default=None)
    plot_offset_checks (dict):
        Only needed for development. Dictionary with kwargs to create plot for checks creating offset regions. If None no plot will be created (default=None)
    plot_offset_regions (dict):
        Only needed for development. Dictionary with kwargs to create plot for offset regions. If None no plot will be created (default=None)
    plot_offset_raster (dict):
        Only needed for development. Dictionary with kwargs to create plot for raster for offset regions. If None no plot will be created (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    
    Returns:
    grid (aabl.Grid):
        a grid covering all points (custom class containing  
    
          
    Examples:
    from aabpl.main import radius_search
    from pandas import read_csv
    pts = read_csv('C:/path/to/file.txt',sep=',',header=None)
    pts.columns = ["eid", "employment", "industry", "lat","lon","moved"]
    grid = radius_search(pts,crs="EPSG:4326",r=750,columns=['employment'])
    grid.plot_vars(filename='employoment_750m')
    """

    (pts, local_crs, columns, x, y, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, help_col
     ) = check_kwargs(
            pts=pts, crs=crs, r=r, columns=columns, x=x, y=y, row_name=row_name,
            col_name=col_name, pts_target=pts_target, x_tgt=x_tgt, y_tgt=y_tgt,
            row_name_tgt=row_name_tgt, col_name_tgt=col_name_tgt, grid=grid, proj_crs=proj_crs, silent=silent,
    )

    # initialize disk_search
    grid.search = DiskSearch(
        grid=grid,
        r=r,
        exclude_pt_itself=exclude_pt_itself,
        include_boundary=include_boundary
    )
    

    # prepare target points data
    grid.search.set_target(
        pts=pts_target,
        columns=columns,
        x=x_tgt,
        y=y_tgt,
        row_name=row_name_tgt,
        col_name=col_name_tgt,
        silent=silent,
    )

    # prepare source points data
    grid.search.set_source(
        pts=pts,
        columns=columns,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        plot_cell_reg_assign=plot_cell_reg_assign,
        plot_offset_checks=plot_offset_checks,
        plot_offset_regions=plot_offset_regions,
        plot_offset_raster=plot_offset_raster,
        silent=silent,
    )
    
    disk_sums_for_pts = grid.search.perform_search(silent=silent,plot_radius_sums=plot_radius_sums,plot_pt_disk=plot_pt_disk)
    if help_col is not None:
        pts_target.drop(columns=[help_col], inplace=True)
    return grid
#

@time_func_perf
def detect_cluster_pts(
    pts:_pd_DataFrame,
    crs:str,
    r:float=0.0075,
    columns:list=[],
    exclude_pt_itself:bool=True,
    k_th_percentiles:float=99.5,
    n_random_points:int=int(1e5),
    random_seed:int=None,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    sum_suffix:str='_750m',
    cluster_suffix:str='_cluster',
    proj_crs:str='auto',
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    row_name_tgt:str=None,
    col_name_tgt:str=None,
    grid:Grid=None,
    include_boundary:bool=False,
    plot_distribution:dict=None,
    plot_radius_sums:dict=None,
    plot_cluster_points:dict=None,
    plot_pt_disk:dict=None,
    plot_cell_reg_assign:dict=None,
    plot_offset_checks:dict=None,
    plot_offset_regions:dict=None,
    plot_offset_raster:dict=None,
    silent:bool=True,
):
    """
    For all points in a DataFrame it searches for all other points (potentially of another DataFrame) within the specified radius and aggregate the values for specified column(s).
    It draws random the bounding box containing all points from DataFrame(s) and aggregate the values within the radius to obtain a random distribution.   
    Then all points from DataFrame which exceed the k_th_percentile of the random distribution are labeld as clustered.
    The results will be appended to DataFrame.

    Args:
    pts (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    columns (list or str):
        column(s) in DataFrame for which data within search radius shall be aggregated. If None provided it will simply count the points within the radius. 
    exclude_pt_itself (bool):
        whether the sums within search radius point shall exlclude the point data itself (default=True)
    k_th_percentiles (float):
        percentile of random distribution that a point needs to exceed to be classified as clustered.
    n_random_points (int):
        number of random points to be drawn to create random distribution (default=100000)
    random_seed (int):
        random seed to be applied when drawing random points to create random distribution. If None no seed will be set (default=None)
    x (str):
        column name of x-coordinate (=longtitude) in pts (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts (default='lat')
    row_name (str):
        name for column that will be appended to pts indicating grid cell row (default='id_y')
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_y')
    sum_suffix (str):
        suffix used for new column(s) creating by aggregating data of columns , 
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    row_name_tgt (str):
        name for column that will be appended to pts indicating grid cell row. If None its assumed to be same as x (default=None)
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_y')
    grid (aabpl.Grid):
        grid of custom class containing points. If None it will automatically one (default=None)
    include_boundary (bool):
        FEATURE NOT YET IMPLEMENTED. Define whether points that are at the distance of exactly the radius shall be considered within (Default=False)
    plot_distribution (dict):
        dictionary with kwargs to create plot for random distribution. If None no plot will be created (default=None)
    plot_radius_sums (dict):
        dictionary with kwargs to create plot for radius sums. If None no plot will be created (default=None)
    plot_pt_disk (dict):
        Only needed for development. Dictionary with kwargs to create plot for example radius search. If None no plot will be created (default=None)
    plot_cell_reg_assign (dict):
        Only needed for development. Dictionary with kwargs to create plot for assginments of points to cell offset regions. If None no plot will be created (default=None)
    plot_offset_checks (dict):
        Only needed for development. Dictionary with kwargs to create plot for checks creating offset regions. If None no plot will be created (default=None)
    plot_offset_regions (dict):
        Only needed for development. Dictionary with kwargs to create plot for offset regions. If None no plot will be created (default=None)
    plot_offset_raster (dict):
        Only needed for development. Dictionary with kwargs to create plot for raster for offset regions. If None no plot will be created (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    
    Returns:
    grid (aabl.Grid):
        a grid covering all points (custom class) with cluster attributes stored to it
    """

    (pts, local_crs, columns, x, y, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, help_col
     ) = check_kwargs(
            pts=pts, crs=crs, r=r, columns=columns, x=x, y=y, row_name=row_name,
            col_name=col_name, pts_target=pts_target, x_tgt=x_tgt, y_tgt=y_tgt,
            row_name_tgt=row_name_tgt, col_name_tgt=col_name_tgt, grid=grid, proj_crs=proj_crs, silent=silent,
    )
    if type(k_th_percentiles) not in [list,_np_array, tuple]:
        k_th_percentiles = [k_th_percentiles for column in columns]
    elif len(k_th_percentiles) < len(columns):
        k_th_percentiles = [k_th_percentiles[i%len(k_th_percentiles)] for i in range(len(columns))]
    # initialize disk_search
    grid.search = DiskSearch(
        grid,
        r=r,
        exclude_pt_itself=exclude_pt_itself,
        include_boundary=include_boundary
    )

    grid.search.set_target(
        pts=pts_target,
        columns=columns,
        x=x_tgt,
        y=y_tgt,
        row_name=row_name_tgt,
        col_name=col_name_tgt,
        silent=silent,
    )

    (cluster_threshold_values, rndm_pts) = get_distribution_for_random_points(
        grid=grid,
        pts=pts,
        columns=columns,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        n_random_points=n_random_points,
        k_th_percentiles=k_th_percentiles,
        plot_distribution=plot_distribution,
        random_seed=random_seed,
        silent=silent,
    )

    if not silent:
        for (colname, threshold_value, k_th_percentile) in zip(columns, cluster_threshold_values,k_th_percentiles):
            print("Threshold value for "+str(k_th_percentile)+"th-percentile is "+str(threshold_value)+" for "+str(colname)+".")
    
    grid.search.set_source(
        pts=pts,
        columns=columns,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        plot_cell_reg_assign=plot_cell_reg_assign,
        plot_offset_checks=plot_offset_checks,
        plot_offset_regions=plot_offset_regions,
        plot_offset_raster=plot_offset_raster,
        silent=silent,
    )


    disk_sums_for_pts = grid.search.perform_search(silent=silent,plot_radius_sums=plot_radius_sums,plot_pt_disk=plot_pt_disk)
    
    # save bool of whether pt is part of a cluster 
    for j, cname in enumerate(columns):
        pts[str(cname)+str(cluster_suffix)] = disk_sums_for_pts.values[:,j]>cluster_threshold_values[j]


    if plot_distribution is not None:
        # print("disk_sums_for_random_points", disk_sums_for_random_points)
        create_distribution_plot(
            pts=pts,
            x=x,
            y=y,
            radius_sum_columns=[n+sum_suffix for n in columns],
            rndm_pts=rndm_pts,
            cluster_threshold_values=cluster_threshold_values,
            k_th_percentiles=k_th_percentiles,
            r=r,
            plot_kwargs=plot_distribution
            )
    #

    def plot_rand_dist(
            pts=pts,
            x=x,
            y=y,
            radius_sum_columns=[n+sum_suffix for n in columns],
            rndm_pts=rndm_pts,
            cluster_threshold_values=cluster_threshold_values,
            k_th_percentiles=k_th_percentiles,
            r=r,
            filename:str="",
            **plot_kwargs
            
    ):
        create_distribution_plot(
            pts=pts,
            x=x,
            y=y,
            radius_sum_columns=radius_sum_columns,
            rndm_pts=rndm_pts,
            cluster_threshold_values=cluster_threshold_values,
            k_th_percentiles=k_th_percentiles,
            r=r,
            filename=filename,
            plot_kwargs=plot_kwargs
            )
    grid.plot.rand_dist = plot_rand_dist
    
    plot_colnames = list(columns) + [n+sum_suffix for n in columns] + [str(cname)+str(cluster_suffix) for cname in columns]
    def plot_cluster_pts(
            self=grid,
            colnames=_np_array(plot_colnames),
            filename:str="",
            **plot_kwargs,
    ):
        return create_plots_for_vars(
            grid=self,
            colnames=colnames,
            filename=filename,
            plot_kwargs=plot_kwargs,
        )
    grid.plot.cluster_pts = plot_cluster_pts

    if plot_cluster_points is not None:
        
        create_plots_for_vars(
            grid=grid,
            colnames=_np_array(plot_colnames),
            plot_kwargs=plot_cluster_points,
        )
        print('create plot for cluster points')
        pass
    
    return grid
# done

def detect_cluster_cells(
    pts:_pd_DataFrame,
    crs:str,
    r:float=750,
    columns:list=[],
    exclude_pt_itself:bool=True,
    k_th_percentiles:float=99.5,
    n_random_points:int=int(1e5),
    random_seed:int=None,
    distance_thresholds:float=2500,
    make_convex:bool=True,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    sum_suffix:str='_750m',
    cluster_suffix:str='_cluster',
    proj_crs:str='auto',
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    row_name_tgt:str=None,
    col_name_tgt:str=None,
    grid:Grid=None,
    include_boundary:bool=False,
    plot_distribution:dict=None,
    plot_radius_sums:dict=None,
    plot_cluster_points:dict=None,
    plot_pt_disk:dict=None,
    plot_cell_reg_assign:dict=None,
    plot_offset_checks:dict=None,
    plot_offset_regions:dict=None,
    plot_offset_raster:dict=None,
    plot_cluster_cells:dict=None,
    silent:bool=True,
):
    """
    For all points in a DataFrame it searches for all other points (potentially of another DataFrame) within the specified radius and aggregate the values for specified column(s).
    It draws random the bounding box containing all points from DataFrame(s) and aggregate the values within the radius to obtain a random distribution.   
    Then all points from DataFrame which exceed the k_th_percentile of the random distribution are labeld as clustered.
    The results will be appended to DataFrame.

    Args:
    pts (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    columns (list or str):
        column(s) in DataFrame for which data within search radius shall be aggregated. If None provided it will simply count the points within the radius. 
    exclude_pt_itself (bool):
        whether the sums within search radius point shall exlclude the point data itself (default=True)
    k_th_percentiles (float):
        percentile of random distribution that a point needs to exceed to be classified as clustered.
    n_random_points (int):
        number of random points to be drawn to create random distribution (default=100000)
    random_seed (int):
        random seed to be applied when drawing random points to create random distribution. If None no seed will be set (default=None)
    x (str):
        column name of x-coordinate (=longtitude) in pts (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts (default='lat')
    row_name (str):
        name for column that will be appended to pts indicating grid cell row (default='id_y')
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_y')
    sum_suffix (str):
        suffix used for new column(s) creating by aggregating data of columns , 
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    row_name_tgt (str):
        name for column that will be appended to pts indicating grid cell row. If None its assumed to be same as x (default=None)
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_y')
    grid (aabpl.Grid):
        grid of custom class containing points. If None it will automatically one (default=None)
    include_boundary (bool):
        FEATURE NOT YET IMPLEMENTED. Define whether points that are at the distance of exactly the radius shall be considered within (Default=False)
    plot_distribution (dict):
        dictionary with kwargs to create plot for random distribution. If None no plot will be created (default=None)
    plot_radius_sums (dict):
        dictionary with kwargs to create plot for radius sums. If None no plot will be created (default=None)
    plot_pt_disk (dict):
        Only needed for development. Dictionary with kwargs to create plot for example radius search. If None no plot will be created (default=None)
    plot_cell_reg_assign (dict):
        Only needed for development. Dictionary with kwargs to create plot for assginments of points to cell offset regions. If None no plot will be created (default=None)
    plot_offset_checks (dict):
        Only needed for development. Dictionary with kwargs to create plot for checks creating offset regions. If None no plot will be created (default=None)
    plot_offset_regions (dict):
        Only needed for development. Dictionary with kwargs to create plot for offset regions. If None no plot will be created (default=None)
    plot_offset_raster (dict):
        Only needed for development. Dictionary with kwargs to create plot for raster for offset regions. If None no plot will be created (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    
    Returns:
    grid (aabl.Grid):
        a grid covering all points (custom class) with cluster attributes stored to it  
    """
    (pts, local_crs, columns, x, y, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, help_col
     ) = check_kwargs(
            pts=pts, crs=crs, r=r, columns=columns, x=x, y=y, row_name=row_name,
            col_name=col_name, pts_target=pts_target, x_tgt=x_tgt, y_tgt=y_tgt,
            row_name_tgt=row_name_tgt, col_name_tgt=col_name_tgt, grid=grid, proj_crs=proj_crs, silent=silent,
    )
    grid = detect_cluster_pts(
        pts=pts,
        crs=local_crs,
        r=r,
        columns=columns,
        exclude_pt_itself=exclude_pt_itself,
        k_th_percentiles=k_th_percentiles,
        n_random_points=n_random_points,
        random_seed=random_seed,
        grid=grid,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        cluster_suffix=cluster_suffix,
        proj_crs=local_crs,
        pts_target=pts_target,
        x_tgt=x_tgt,
        y_tgt=y_tgt,
        row_name_tgt=row_name_tgt,
        col_name_tgt=col_name_tgt,
        include_boundary=include_boundary,
        plot_distribution=plot_distribution,
        plot_radius_sums=plot_radius_sums,
        plot_cluster_points=plot_cluster_points,
        plot_pt_disk=plot_pt_disk,
        plot_cell_reg_assign=plot_cell_reg_assign,
        plot_offset_checks=plot_offset_checks,
        plot_offset_regions=plot_offset_regions,
        plot_offset_raster=plot_offset_raster,
        silent=silent,
    )
    
    grid.clustering.create_clusters(
        pts=pts,
        columns=columns,
        distance_thresholds=distance_thresholds,
        make_convex=make_convex,
        row_name=row_name,
        col_name=col_name,
        cluster_suffix=cluster_suffix,
        plot_cluster_cells=plot_cluster_cells,
        )
    
    return grid
#


#!python
#cython: wraparound=False, nonecheck=False, boundscheck=False, cdivision=True, language_level=3
from numpy cimport npy_intp

from libc.stdint cimport (uint64_t, int32_t, int64_t)
from numpy.random cimport bitgen_t

cdef extern from "numpy/random/distributions.h":

    struct s_binomial_t:
        int has_binomial
        double psave
        int64_t nsave
        double r
        double q
        double fm
        int64_t m
        double p1
        double xm
        double xl
        double xr
        double c
        double laml
        double lamr
        double p2
        double p3
        double p4

    ctypedef s_binomial_t binomial_t

    float random_standard_uniform_f(bitgen_t *bitgen_state) nogil
    double random_standard_uniform(bitgen_t *bitgen_state) nogil
    void random_standard_uniform_fill(bitgen_t* bitgen_state, npy_intp cnt, double *out) nogil
    void random_standard_uniform_fill_f(bitgen_t *bitgen_state, npy_intp cnt, float *out) nogil
    
    double random_standard_exponential(bitgen_t *bitgen_state) nogil
    float random_standard_exponential_f(bitgen_t *bitgen_state) nogil
    void random_standard_exponential_fill(bitgen_t *bitgen_state, npy_intp cnt, double *out) nogil
    void random_standard_exponential_fill_f(bitgen_t *bitgen_state, npy_intp cnt, float *out) nogil
    void random_standard_exponential_inv_fill(bitgen_t *bitgen_state, npy_intp cnt, double *out) nogil
    void random_standard_exponential_inv_fill_f(bitgen_t *bitgen_state, npy_intp cnt, float *out) nogil
    
    double random_standard_normal(bitgen_t* bitgen_state) nogil
    float random_standard_normal_f(bitgen_t *bitgen_state) nogil
    void random_standard_normal_fill(bitgen_t *bitgen_state, npy_intp count, double *out) nogil
    void random_standard_normal_fill_f(bitgen_t *bitgen_state, npy_intp count, float *out) nogil
    double random_standard_gamma(bitgen_t *bitgen_state, double shape) nogil
    float random_standard_gamma_f(bitgen_t *bitgen_state, float shape) nogil

    float random_standard_uniform_f(bitgen_t *bitgen_state) nogil
    void random_standard_uniform_fill_f(bitgen_t* bitgen_state, npy_intp cnt, float *out) nogil
    float random_standard_normal_f(bitgen_t* bitgen_state) nogil
    float random_standard_gamma_f(bitgen_t *bitgen_state, float shape) nogil

    int64_t random_positive_int64(bitgen_t *bitgen_state) nogil
    int32_t random_positive_int32(bitgen_t *bitgen_state) nogil
    int64_t random_positive_int(bitgen_t *bitgen_state) nogil
    uint64_t random_uint(bitgen_t *bitgen_state) nogil

    double random_normal(bitgen_t *bitgen_state, double loc, double scale) nogil

    double random_gamma(bitgen_t *bitgen_state, double shape, double scale) nogil
    float random_gamma_f(bitgen_t *bitgen_state, float shape, float scale) nogil

    double random_exponential(bitgen_t *bitgen_state, double scale) nogil
    double random_uniform(bitgen_t *bitgen_state, double lower, double range) nogil
    double random_beta(bitgen_t *bitgen_state, double a, double b) nogil
    double random_chisquare(bitgen_t *bitgen_state, double df) nogil
    double random_f(bitgen_t *bitgen_state, double dfnum, double dfden) nogil
    double random_standard_cauchy(bitgen_t *bitgen_state) nogil
    double random_pareto(bitgen_t *bitgen_state, double a) nogil
    double random_weibull(bitgen_t *bitgen_state, double a) nogil
    double random_power(bitgen_t *bitgen_state, double a) nogil
    double random_laplace(bitgen_t *bitgen_state, double loc, double scale) nogil
    double random_gumbel(bitgen_t *bitgen_state, double loc, double scale) nogil
    double random_logistic(bitgen_t *bitgen_state, double loc, double scale) nogil
    double random_lognormal(bitgen_t *bitgen_state, double mean, double sigma) nogil
    double random_rayleigh(bitgen_t *bitgen_state, double mode) nogil
    double random_standard_t(bitgen_t *bitgen_state, double df) nogil
    double random_noncentral_chisquare(bitgen_t *bitgen_state, double df,
                                       double nonc) nogil
    double random_noncentral_f(bitgen_t *bitgen_state, double dfnum,
                               double dfden, double nonc) nogil
    double random_wald(bitgen_t *bitgen_state, double mean, double scale) nogil
    double random_vonmises(bitgen_t *bitgen_state, double mu, double kappa) nogil
    double random_triangular(bitgen_t *bitgen_state, double left, double mode,
                             double right) nogil

    int64_t random_poisson(bitgen_t *bitgen_state, double lam) nogil
    int64_t random_negative_binomial(bitgen_t *bitgen_state, double n, double p) nogil
    int64_t random_binomial(bitgen_t *bitgen_state, double p, int64_t n, binomial_t *binomial) nogil
    int64_t random_logseries(bitgen_t *bitgen_state, double p) nogil
    int64_t random_geometric_search(bitgen_t *bitgen_state, double p) nogil
    int64_t random_geometric_inversion(bitgen_t *bitgen_state, double p) nogil
    int64_t random_geometric(bitgen_t *bitgen_state, double p) nogil
    int64_t random_zipf(bitgen_t *bitgen_state, double a) nogil
    int64_t random_hypergeometric(bitgen_t *bitgen_state, int64_t good, int64_t bad,
                                    int64_t sample) nogil

    uint64_t random_interval(bitgen_t *bitgen_state, uint64_t max) nogil

    # Generate random uint64 numbers in closed interval [off, off + rng].
    uint64_t random_bounded_uint64(bitgen_t *bitgen_state,
                                   uint64_t off, uint64_t rng,
                                   uint64_t mask, bint use_masked) nogil

    void random_multinomial(bitgen_t *bitgen_state, int64_t n, int64_t *mnix,
                            double *pix, npy_intp d, binomial_t *binomial) nogil

    int random_multivariate_hypergeometric_count(bitgen_t *bitgen_state,
                          int64_t total,
                          size_t num_colors, int64_t *colors,
                          int64_t nsample,
                          size_t num_variates, int64_t *variates) nogil
    void random_multivariate_hypergeometric_marginals(bitgen_t *bitgen_state,
                               int64_t total,
                               size_t num_colors, int64_t *colors,
                               int64_t nsample,
                               size_t num_variates, int64_t *variates) nogil




# intersection of two circles with same radius
from numpy import array as _np_array, sign as _np_sign, arange as _np_arange, invert as _np_invert, zeros as _np_zeros
from numpy.linalg import norm as _np_linalg_norm
from math import log10 as _math_log10
from matplotlib import pyplot as plt
from aabpl.utils.general import make_bins_from_vals, get_vals_from_bins
from aabpl.utils.distances_to_cell import (get_cell_closest_point_to_point, get_cell_farthest_vertex_to_point,
 check_if_never_contains_convex_set, check_if_always_overlaps_full_convex_set, get_cells_relevant_for_disk_by_type)
from .offset_region_classes import OffsetRegion, Vertex, LineSegment, Circle
from aabpl.illustrations.illustrate_cell_pattern import plot_cell_pattern
from aabpl.testing.test_performance import time_func_perf
from shapely.geometry import Polygon, LineString, Point
from geopandas import GeoSeries, GeoDataFrame


def create_triangle_1_region(
        clear_all:bool=True,
        convex_set_coordiantes:list=[(0.,0.), (0.5,0.), (0.5,0.5)]
) -> dict:
    """
    Delete all Regions, Edges, Vertices from dicts. Create new vertices, edges and region for triangle 1
    """
    if clear_all: 
        all_regions = dict()
        all_edges = dict()
        all_vtx = dict()
    
    vertices_set = [Vertex(x=x, y=y, all_vtx=all_vtx) for (x,y) in convex_set_coordiantes]
    edges_set = [LineSegment(vtx1=v1, vtx2=v2, all_edges=all_edges) for v1,v2 in zip(vertices_set, vertices_set[1:]+vertices_set[:1])]
    OffsetRegion(edges=edges_set, checks=[], all_regions=all_regions)
    
    return all_regions
#


def add_circle_check_to_dict(
        cell:_np_array,
        nev_cn:bool,
        alw_ov:bool,
        check_dict,
        r:float,
        grid_spacing:float,
):
    """
    TODO 
    updates check dicts with check(s)
    """
    trgl_pt = (0.25,0.125)
    point_in_triangle1 = _np_array([(0.25,0.125)]) # TODO remove wraping list
    
    if not alw_ov:
        closest_pt = tuple([float(v) for v in get_cell_closest_point_to_point(trgl_pt, cell)])
        if closest_pt not in check_dict:
            check_dict[closest_pt] = {'split_edge': Circle(center=closest_pt, r=r/grid_spacing)}
        check_dict[closest_pt]['overlaps'] = cell
        check_dict[closest_pt]['split_edge'].overlaps = tuple([*cell])
    #

    if not nev_cn:
        farthest_pt = tuple([float(v) for v in get_cell_farthest_vertex_to_point(point_in_triangle1, cell)[0]])
        if farthest_pt not in check_dict:
            check_dict[farthest_pt] = {'split_edge': Circle(center=farthest_pt, r=r/grid_spacing)}

        check_dict[farthest_pt]['contains'] = cell
        check_dict[farthest_pt]['split_edge'].contains = tuple([*cell])
    #
#

def add_line_check_to_dict(
        cell:_np_array,
        nev_cn:bool,
        alw_ov:bool,
        check_dict,
        all_regions:dict,
        r:float,
        grid_spacing:float,
):
    """
    TODO maybe already add more check logic here e.g the segment
    updates check dicts with check
    """
        
    row, col = [int(c) for c in cell]
    
    an_edge  = list(all_regions.values())[-1].edges[-1]
    all_edges = an_edge.all_edges
    all_vtx = an_edge.vtx1.all_vtx
    
    if row == 0 and col == 0:
        raise ValueError('Not implemented. Choose grid spacing s.t. search radius**2 > (2*grid_spacing**2).')    
    
    if not alw_ov:
        if col == 0.:
            split_edge = LineSegment(
                vtx1=Vertex(0.0 if row > 0 else 0.5,  (row - (.5 - r) * _np_sign(row)), all_vtx),
                vtx2=Vertex(0.0 if row < 0 else 0.5,  (row - (.5 - r) * _np_sign(row)), all_vtx), 
                all_edges=all_edges
            )
        else:
            split_edge = LineSegment(
                vtx1=Vertex((col - (.5 - r) * _np_sign(col)), 0.5 if col > 0 else 0.0, all_vtx),
                vtx2=Vertex((col - (.5 - r) * _np_sign(col)), 0.5 if col < 0 else 0.0, all_vtx), 
                all_edges=all_edges
            )
        split_edge.overlaps = (row, col)
        check_dict[split_edge] = {'split_edge': split_edge, 'overlaps': (row,col)}
    
    if not nev_cn:
        farthest_pt = (
            -0.5 if col == 0 else col + .5 * _np_sign(col),
            -0.5 if row == 0 else row + .5 * _np_sign(row)
        )
        if farthest_pt not in check_dict:
            check_dict[farthest_pt] = {'split_edge': Circle(center=farthest_pt, r=r/grid_spacing)}
        
        check_dict[farthest_pt]['split_edge'].contains = (row, col)
        check_dict[farthest_pt]['contains'] = cell
        # # THESE RESULTS WILL BE CHECKED ANYWAYS. AT THE END YOU CAN REQUEST THOSE RESULTS
        # check_dict[(x,y)] = {'cells_to_overlap': cells_to_overlap, 'contains': (x,y)}
    #
#

def create_check_dict(
    cells_to_check,
    all_regions:dict,
    r:float=3,
    grid_spacing:float=1,
    include_boundary: bool = False        
):
    """
    Gets all cells that are potentially overlap or conain or contain set if buffered by radius
    TODO ensure r/grid_spacing covers all cases
    """
    check_dict = dict()
    triangle_1_vertices = _np_array([[0, 0], [0.5, 0], [0.5, 0.5]])

    cells_always_overlapped = check_if_always_overlaps_full_convex_set(
        cells=cells_to_check,
        convex_set_vertices=triangle_1_vertices,
        r=r,
        grid_spacing=grid_spacing,
        vertex_is_inside_convex_set=True,
        include_boundary=include_boundary,
    )

    cells_never_contained = check_if_never_contains_convex_set(
        cells=cells_to_check,
        convex_set_vertices=triangle_1_vertices,
        r=r,
        grid_spacing=grid_spacing,
        vertex_is_inside_convex_set=True,
        include_boundary=include_boundary,
    )

    cells_alw_only_overlapped = []
    for cell, alw_ov, nev_cn in zip(cells_to_check, cells_always_overlapped, cells_never_contained):
        if alw_ov and nev_cn:
            cells_alw_only_overlapped.append(cell) # TODO this can be removed - no longer necessary to store those.
        if 0 in cell: # cell in same column or row
            add_line_check_to_dict(cell=cell, nev_cn=nev_cn, alw_ov=alw_ov, check_dict=check_dict, all_regions=all_regions, r=r, grid_spacing=grid_spacing,)
        else:
            add_circle_check_to_dict(cell=cell, nev_cn=nev_cn, alw_ov=alw_ov, check_dict=check_dict, r=r, grid_spacing=grid_spacing,)
        #
    #
    return check_dict, cells_to_check[cells_always_overlapped]
#

def apply_checks_to_create_regions(
        check_dict,
        trgl_regions:dict,
        r:float=3,
        plot_offset_checks:dict=None,
        axs = None
    ):
    """
    TODO get check(s) plural / singular clean.
    """

    if not plot_offset_checks is None:
        if axs is None:
            nrows = int(len(check_dict)**.5)
            ncols = -int(-len(check_dict.items())//nrows) 
            fig,axs = plt.subplots(nrows, ncols, figsize=(ncols*5,nrows*5))

    for i, (key, check) in enumerate(check_dict.items()):
        colors=[]
        split_edge = check['split_edge']
        regions = list(trgl_regions.values())
        for region in regions:
            # check if any pt is within readius
            colors.append(region.split_with_circle(split_edge, check))

        if not plot_offset_checks is None:
            ax = axs.flat[i]
            OffsetRegion.plot_many(regions=regions, ax=ax, color=colors, alpha=0.8, plot_edges=False)
            # OffsetRegion.plot_many(regions=list(trgl_regions.values()), ax=ax, color=colors, alpha=0.8, plot_edges=False)
            if split_edge.type != 'LineSegment':
                x,y = [float(c) for c in key]
                split_edge = check['split_edge']
                GeoSeries([Point((x,y)).buffer(r,30)]).plot(ax=ax, color='None', edgecolor='blue')
                ax.set_title(str((x,y))+ "center.")
            else:
                GeoSeries([LineString([split_edge.vtx1.xy, split_edge.vtx2.xy])]).plot(ax=ax, color='None', edgecolor='red')
    

    
    # Check if all checks are performaned on each, else throw error 
    n_checks = [len(region.checks) for region in list(trgl_regions.values())]
    if not n_checks.count(n_checks[0]) == len(n_checks):
        raise ValueError("The number of checks performed differ among micro regions. They are expected to be all of the same Length.", n_checks)
    
    

def cleanup_region_check_results(
        trgl_regions:dict, 
        cells_contained_in_all_trgl_disks:_np_array,
        cells_always_overlapped:list,
        all_cells:_np_array,
        grid_spacing:float,
        r:float, 
        plot_offset_regions:dict=None):
    # now all checks are added to regions
    # ensure that each region.checks has the same length!
    if not plot_offset_regions is None:
        fig, axs = plt.subplots(nrows=len(trgl_regions), ncols=2, figsize=(8, 4*len(trgl_regions)))
    for n, region in enumerate(list(trgl_regions.values())):
        region.contained_cells = []
        region.overlapped_cells = []
        for check in region.checks:
            if not check['result'] == True:
                continue
            if 'contains' in check:
                region.contained_cells.append(check['contains'])
            if 'overlaps' in check:
                region.overlapped_cells.append(check['overlaps'])

        # add all cells contained in cell for triangle 1 (not including cells that are contained for any pt inside cell)
        region.contained_cells = tuple(sorted(set(
            [(x,y) for x,y in cells_contained_in_all_trgl_disks] + 
            [(x,y) for (x,y) in region.contained_cells]
        )))

        # add all cells that are always at least overlapped and not contained in this one
        region.overlapped_cells = tuple(sorted(set(
            [(x,y) for (x,y) in cells_always_overlapped if not ((x,y) in region.contained_cells)] + 
            [(x,y) for (x,y) in region.overlapped_cells]
        )))

        region.shared_along_vert = any([edge.vtx1.y==0 and edge.vtx2.y==0  for edge in region.edges])
        region.shared_along_diag = any([edge.vtx1.x==edge.vtx1.y and edge.vtx2.x==edge.vtx2.y for edge in region.edges])

        if not plot_offset_regions is None:
            region.plot_many(regions=list(trgl_regions.values()), ax=axs.flat[n*2], alpha=0.1, add_idxs=False)
            region.plot_single(ax=axs.flat[n*2], alpha=1, add_idx_edges=False)
            region.plot_single(ax=axs.flat[n*2+1], alpha=1, plot_edges=False, add_idx_edges=False)
            plot_cell_pattern(
                contained_cells=region.contained_cells,
                overlapped_cells=region.overlapped_cells,
                all_cells=all_cells,
                ax=axs.flat[n*2 + 1],
                r=r,
                grid_spacing=grid_spacing,
            )
    #


def transform_region_to_remaining_triangles(trgl_regions):
    regions = list(trgl_regions.values())
    unique_contained_cells = dict()
    unique_overlapped_cells = dict()
    translate_trgl_reg_nr_to_reg_nr = dict()
    # create new regions for rotation if not similar
    condensed_regions = []
    all_regions = []
    
    for nr, region in enumerate(regions):
        region.nr = nr*10+1
        new_regions = []
        for i in [1,2,3,4,5,6,7,8]:
                # create new region!
                rotated_region = region.transform_to_trgl(i)
                rotated_region.nr = nr*10+i
                new_regions.append(rotated_region)
        
        all_regions.extend(new_regions)
        
        if region.shared_along_vert and region.shared_along_diag:
            condensed_regions.append(OffsetRegion.merge_regions(new_regions))
            translate_trgl_reg_nr_to_reg_nr.update({nr*10+j: nr*10+1 for j in [1,2,3,4,5,6,7,8]})
            pass
        elif region.shared_along_vert:
            condensed_regions.append(OffsetRegion.merge_regions([new_regions[-1], new_regions[0]]))
            condensed_regions.append(OffsetRegion.merge_regions(new_regions[1:3]))
            condensed_regions.append(OffsetRegion.merge_regions(new_regions[3:5]))
            condensed_regions.append(OffsetRegion.merge_regions(new_regions[5:7]))
            translate_trgl_reg_nr_to_reg_nr.update({
                nr*10+8: nr*10+1, nr*10+1: nr*10+1,
                nr*10+2: nr*10+3, nr*10+3: nr*10+3,
                nr*10+4: nr*10+5, nr*10+5: nr*10+5,
                nr*10+6: nr*10+7, nr*10+7: nr*10+7,
                })
            pass
        elif region.shared_along_diag:
            condensed_regions.append(OffsetRegion.merge_regions(new_regions[0:2]))
            condensed_regions.append(OffsetRegion.merge_regions(new_regions[2:4]))
            condensed_regions.append(OffsetRegion.merge_regions(new_regions[4:6]))
            condensed_regions.append(OffsetRegion.merge_regions(new_regions[6:8]))
            translate_trgl_reg_nr_to_reg_nr.update({
                nr*10+1: nr*10+1, nr*10+2: nr*10+1,
                nr*10+3: nr*10+3, nr*10+4: nr*10+3,
                nr*10+5: nr*10+5, nr*10+6: nr*10+5,
                nr*10+7: nr*10+7, nr*10+8: nr*10+7,
                })
            pass
        else:
            condensed_regions.extend(new_regions)
            translate_trgl_reg_nr_to_reg_nr.update({nr*10+j: nr*10+j for j in [1,2,3,4,5,6,7,8]})

    all_regions = condensed_regions
    trgl_reg_nr_to_id = dict() 
    all_regions.sort(key=lambda reg: (reg.trgl_nr))
    
    for region in all_regions:
        if not region.contained_cells in unique_contained_cells:
            unique_contained_cells[region.contained_cells] = len(unique_contained_cells)
        if not region.overlapped_cells in unique_overlapped_cells:
            unique_overlapped_cells[region.overlapped_cells] = len(unique_overlapped_cells)
    
    contain_region_mult = 10**(int(_math_log10(len(unique_overlapped_cells)))+1)
    id_to_offset_regions = dict()
    
    for region in all_regions:
        region.id = unique_contained_cells[region.contained_cells] * contain_region_mult + unique_overlapped_cells[region.overlapped_cells]
        
        if not region.id in id_to_offset_regions:
            id_to_offset_regions[region.id] = region
        
        for i in range(0, 8 + 1 - region.trgl_nr):
            trgl_reg_nr_to_id[region.nr + i] = region.id
    
    translate_reg_nr_to_reg_id = {k: trgl_reg_nr_to_id[v] for k,v in translate_trgl_reg_nr_to_reg_nr.items()}

    return id_to_offset_regions, translate_reg_nr_to_reg_id, contain_region_mult
    
#

# TODO ensure that row,col x,y is not mixed up

def assign_pts_to_offset_region(
        pts,
        potential_regions,
):
    [[pt for reg in potential_regions] for pt in pts]
    pass


def create_raster_plot(
        regions:list,
        raster_cell_to_regions:dict,
        offset_x_bins,
        offset_y_bins,
        lims=[-0.05, 0.55],
        add_raster_labels:bool=True
        
        ):
        fig, axs = plt.subplots(1,3,figsize=(50,30))

        gdf = GeoDataFrame(geometry=[Polygon([(_np_sign(ix)*x, _np_sign(iy)*y) for x,y in [
            (offset_x_bins[abs(ix)-1][0], offset_y_bins[abs(iy)-1][0]), (offset_x_bins[abs(ix)-1][1], offset_y_bins[abs(iy)-1][0]),
            (offset_x_bins[abs(ix)-1][1], offset_y_bins[abs(iy)-1][1]), (offset_x_bins[abs(ix)-1][0], offset_y_bins[abs(iy)-1][1])]]
        ) for (ix, iy) in raster_cell_to_regions])
        lens = [len(v) for v in raster_cell_to_regions.values()]
        gdf['rel'] = [(l/(max(lens)), l/(max(lens))/2,1- l/(max(lens))) for l in lens]
        gdf.plot(ax=axs.flat[0],  color=gdf['rel'], edgecolor='None', alpha=0.7, legend=True, linewidth=0.15)
        
        gdf.plot(ax=axs.flat[1],  color=gdf['rel'], edgecolor='None', alpha=0.7, legend=True, linewidth=0.15)
        region_comb_dict = dict()
        for l, region_comb in sorted([(len(v), tuple([r.id for r in v])) for v in raster_cell_to_regions.values()]):
            if not region_comb in region_comb_dict:
                region_comb_dict[region_comb] = len(region_comb_dict)
        region_comb_nrs = [region_comb_dict[tuple([r.id for r in v])] for v in raster_cell_to_regions.values()]
        gdf.plot(ax=axs.flat[2],  column=_np_array(region_comb_nrs), cmap='viridis', edgecolor='None', alpha=0.7, linewidth=0.15)
        
        if add_raster_labels:
            for (ix,iy), v in raster_cell_to_regions.items():
                (x_low, x_up), (y_low, y_up) = [_np_sign(ix)*x for x in offset_x_bins[abs(ix)-1]], [_np_sign(iy)*y for y in offset_y_bins[abs(iy)-1]]
                # axs.flat[1].annotate(text=".".join([str(v0) for v0 in v]), xy=((x_low+x_up)/2, (y_low+y_up)/2), horizontalalignment='center', fontsize=5)
                # axs.flat[2].annotate(text=".".join([str(v0) for v0 in v]), xy=((x_low+x_up)/2, (y_low+y_up)/2), horizontalalignment='center', fontsize=5)
                axs.flat[1].annotate(text=str(len(v)), xy=((x_low+x_up)/2, (y_low+y_up)/2), horizontalalignment='center', fontsize=8)
                axs.flat[2].annotate(text=str(region_comb_dict[tuple([r.id for r in v])]), xy=((x_low+x_up)/2, (y_low+y_up)/2), horizontalalignment='center', fontsize=8)
        else:
            for (ix,iy), v in raster_cell_to_regions.items():
                (x_low, x_up), (y_low, y_up) = [_np_sign(ix)*x for x in offset_x_bins[abs(ix)-1]], [_np_sign(iy)*y for y in offset_y_bins[abs(iy)-1]]
                text = str(len(v))#str(ix)#+"."+str(iy)
                # text = "+"
                axs.flat[1].annotate(text=text, xy=((x_low+x_up)/2, (y_low+y_up)/2), horizontalalignment='center', fontsize=7)
        
        OffsetRegion.plot_many(regions=regions, plot_edges=False, edgecolor='black', ax=axs.flat[0], facecolor='None', alpha=0.8, linewidth=0.4)
        OffsetRegion.plot_many(regions=regions, plot_edges=False, edgecolor='black', ax=axs.flat[1], facecolor='None', alpha=0.8, linewidth=0.4, add_idxs=False)
        OffsetRegion.plot_many(regions=regions, plot_edges=False, edgecolor='black', ax=axs.flat[2], facecolor='None', alpha=0.8, linewidth=0.4, add_idxs=False)
        for ax in axs.flat:
            ax.set_xlim(lims)
            ax.set_ylim(lims)
        axs.flat[-1].legend((i for i in sorted(set(lens))))


def sort_trgl_region_into_raster(
    trgl_regions:dict,
    vertices:list=None,
    plot_offset_raster:dict=None,
):
    """
    TODO CLEAN UP
    
    """
    # assign each cell to a region
    # start with triangle only
    regions_to_check = [reg for reg in trgl_regions.values() if reg.trgl_nr in [1,2]]
    regions_to_check.sort(key=lambda reg: (reg.xmin, reg.xmax, reg.ymin, reg.ymax))
    
    if vertices is None:
        vertices = []
        for region in regions_to_check:
            if region.xmin >= region.xmax:
                print("xmin>=xmax", region.coords)
            vertices.extend(region.vertices)
    
    if False:
        unique_x_vals = sorted(set([vtx.x for vtx in vertices if vtx.x >= 0]))
        unique_y_vals = sorted(set([vtx.y for vtx in vertices if vtx.y >= 0]))
        make_bins_from_vals, get_vals_from_bins
        offset_x_bins = make_bins_from_vals(unique_x_vals) 
        offset_y_bins = make_bins_from_vals(unique_y_vals) 
    else:
        unique_x_or_y_vals = sorted(set([vtx.x for vtx in vertices if vtx.x >= 0] + [vtx.y for vtx in vertices if vtx.y >= 0]))
        offset_x_bins = make_bins_from_vals(unique_x_or_y_vals)
        offset_y_bins = make_bins_from_vals(unique_x_or_y_vals)
        
    
    # remember regions are not always convex - but maybe i guess they will be once splitted along lines
    raster_cell_to_regions = dict()
    unique_reg_id_combs_in_raster_cells = dict()
    i, j = 0, 0
    regions_to_check_at_x = regions_to_check
    for ix, (x_low, x_up) in zip(range(1, len(offset_x_bins)+1), offset_x_bins):
        
        # look for leftmost region that overlaps x_low  
        i = next((ix for ix, reg in enumerate(regions_to_check_at_x) if reg.xmax > x_low),-1)
        regions_to_check_at_x = regions_to_check_at_x[i:]
        
        if len(regions_to_check_at_x)==0:
            print("Break, no regions to check", x_low, x_up )
            break
        if j != -1:
            j = next((jx for jx, reg in enumerate(regions_to_check_at_x) if reg.xmin >= x_up),-1)
        else:
            print("-------------- j != -1 --------------")

        regions_to_check_at_xy = regions_to_check_at_x[:j]
        for iy, (y_low, y_up) in zip(range(1, len(offset_y_bins)+1), offset_y_bins):
            if y_low >= x_up: 
                break
            # look for downmost region that overlaps y_low  
            n = len(regions_to_check_at_xy)
            regions_to_check_at_xy = [reg for reg in regions_to_check_at_xy if reg.ymax > y_low]
            
            # TODO leverage this performance gain later on. 
            # # extract all regions that within raster_cell
            # def check_if_raster_cell_overlaps_region(reg:OffsetRegion,x_low:float, y_low:float, x_up:float, y_up:float):
            #     raster_vertices = ((x_low, y_low), (x_up, y_low), (x_up, y_up), (x_low, y_up))
            #     for (x,y) in list(reg.get_coords()):
            #         for vx,vy in raster_vertices:
            #             if (vx)
            #     return False
            regions_at_raster_cell = []
            # region_nrs_at_raster_cell = set()
            region_ids_at_raster_cell = set()
            for reg in regions_to_check_at_xy:
                if reg.ymin < y_up and not reg.id in region_ids_at_raster_cell:
                    # region_nrs_at_raster_cell.add(reg.nr)
                    region_ids_at_raster_cell.add(reg.id)
                    regions_at_raster_cell.append(reg)

            # region_nrs_at_raster_cell = tuple(sorted(region_nrs_at_raster_cell))
            # if len(region_nrs_at_raster_cell)==0:
            #     print("(x_low, x_up)", (x_low, x_up), "(y_low, y_up)", (y_low, y_up), 'N', n, 'n', len(regions_to_check_at_xy), 'Nx', len(regions_to_check_at_x))
            
            region_ids_at_raster_cell = tuple(sorted(region_ids_at_raster_cell))
            
            if not region_ids_at_raster_cell in unique_reg_id_combs_in_raster_cells:
                unique_reg_id_combs_in_raster_cells[region_ids_at_raster_cell] = []
            
            unique_reg_id_combs_in_raster_cells[region_ids_at_raster_cell].append((ix,iy))
            raster_cell_to_regions[(ix, iy)] = regions_at_raster_cell
        #
    #

    # print("n cells in triangle:", len(raster_cell_to_regions))
    # print("mean regions at raster cell", sum([len(x) for x in raster_cell_to_regions.values()])/len(raster_cell_to_regions))
    # print("unique_reg_id_combs_in_raster_cells", unique_reg_id_combs_in_raster_cells)
    if not plot_offset_raster is None:
        create_raster_plot(regions=list(trgl_regions.values()), raster_cell_to_regions=raster_cell_to_regions, offset_x_bins=offset_x_bins, offset_y_bins=offset_y_bins)
    
    return raster_cell_to_regions, offset_x_bins, offset_y_bins, unique_reg_id_combs_in_raster_cells
#

def create_radius_check(edge, r:float, include_boundary:bool):
    (x,y) = edge.center
    if include_boundary:
        return lambda pts: _np_linalg_norm(pts - (x,y), axis=1) <= r
    return lambda pts: _np_linalg_norm(pts - (x,y), axis=1) < r
#

def create_line_check(edge, r:float, include_boundary:bool):
    col_index = int(edge.vtx1.y == edge.vtx2.y)
    val = edge.vtx1.xy[col_index]
    if include_boundary:
        return lambda pts: abs(pts[:,col_index] - val) <= r
    return lambda pts: abs(pts[:,col_index] - val) < r

#

def create_pt_checks(edge, r:float, include_boundary:bool=False):
    if edge.type == 'Arc':
        return create_radius_check(edge, r, include_boundary)
    return create_line_check(edge, r, include_boundary)
#

def edge_is_shared_with_region_id(
        reversed_edge_coords,
        remaining_regions,
):
    for reg in remaining_regions:
        for e in reg.edges:
            if reversed_edge_coords == e.coords:
                return reg.id
    return None
#

def edge_is_shared_with_region_id2(
        edge,
        remaining_regions,
):
    attr_to_check = 'contains' if hasattr(edge, 'contains') else 'overlaps'
    if not hasattr(edge, attr_to_check):
        return None

    for reg in remaining_regions:
        for e in reg.edges:
            if hasattr(e, attr_to_check) and getattr(edge, attr_to_check) == getattr(e, attr_to_check):
                return reg.id
    return None
#

# create (potentially recursive checks:)
def add_check_to_tree_at_pos(
        tree_pos,
        checks,
        remaining_regions
):
    edge, check = checks[0]
    tree_pos['check'] = check
    if len(remaining_regions)<=1:
        raise ValueError("remaining_regions", len(remaining_regions), len(checks))
    regions_if_true = [reg for reg in remaining_regions if 
                       (hasattr(edge, 'contains') and edge.contains in reg.contained_cells) or 
                       (hasattr(edge, 'overlaps') and edge.overlaps in reg.overlapped_cells)]
    ids_if_true = [reg.id for reg in regions_if_true]
    regions_if_false = [reg for reg in remaining_regions if reg.id not in ids_if_true]
    if len(regions_if_true) == 0:
        pass
    elif len(regions_if_true) == 1:
        id_if_true = regions_if_true[0].id
        tree_pos[True] = id_if_true
    else:
        if len(regions_if_false) > 0:
            tree_pos[True] = {}
            add_check_to_tree_at_pos(tree_pos[True], checks[1:], regions_if_true)
        else:
            add_check_to_tree_at_pos(tree_pos, checks[1:], regions_if_true)

    if len(regions_if_false) == 0:
        pass
    elif len(regions_if_false)==1:
        id_if_false = regions_if_false[0].id
        tree_pos[False] = id_if_false
    else:
        if len(regions_if_true) > 0:
            tree_pos[False] = {}
            add_check_to_tree_at_pos(tree_pos[False], checks[1:], regions_if_false)
        else:
            add_check_to_tree_at_pos(tree_pos, checks[1:], regions_if_false)

    if len(regions_if_true)+len(regions_if_false) == 0:
        raise ValueError('remaining_regions', remaining_regions, 'checks', checks)
#

def determine_offset_region_for_pts_inner(
        pts:_np_array,
        check_tree
    ):

    if type(check_tree) != dict:
        return _np_zeros(len(pts), int) + check_tree
    res = _np_zeros(len(pts), int)
    check_res = check_tree['check'](pts)
    res[_np_arange(len(pts))[check_res]] = determine_offset_region_for_pts_inner(pts=pts[_np_arange(len(pts))[check_res]], check_tree=check_tree[True])
    res[_np_arange(len(pts))[_np_invert(check_res)]] = determine_offset_region_for_pts_inner(pts=pts[_np_arange(len(pts))[_np_invert(check_res)]], check_tree=check_tree[False])
    
    return res
#

def prepare_raster_to_offset_region_checks(
        id_to_offset_regions,
        unique_reg_id_combs_in_raster_cells,
        r, 
        include_boundary,
    ):
    
    offset_reg_id_comb_to_check = dict()

    for region_comb in unique_reg_id_combs_in_raster_cells:
        regions = [id_to_offset_regions[id] for id in sorted(set(region_comb))]

        if len(regions)==0:
            raise ValueError("PP", region_comb)
        if len(region_comb) == 1:
            region_id = regions[0].id
            offset_reg_id_comb_to_check[region_comb] = lambda pts, region_id=region_id: _np_zeros(len(pts),int)+region_id
            continue
        
        check_tree = dict()
        check_edges_for_regions = []
        for i, region in enumerate(regions):
            for edge in region.edges:
                # reversed_coords = (edge.coords[1], edge.coords[0])
                # shared_with_reg_id = edge_is_shared_with_region_id(reversed_coords, regions[i+1:])
                shared_with_reg_id = edge_is_shared_with_region_id2(edge, regions[i+1:])
                if not shared_with_reg_id is None:
                    edge_check = create_pt_checks(edge, r=r, include_boundary=include_boundary)
                    check_edges_for_regions.append((edge, edge_check))
        
        # print("len(check_edges_for_regions)", len(check_edges_for_regions), len(regions), set([reg.trgl_nr for reg in regions]))
        try:
            add_check_to_tree_at_pos(tree_pos=check_tree, checks=check_edges_for_regions, remaining_regions=regions)
            pass
        except:
            OffsetRegion.plot_many(regions=regions, plot_vertices=False, x_lim=(-.52,.52), y_lim=(-.52,.52))
            raise ValueError("trgl_nrs", [(reg.trgl_nr, reg.id, reg.trgl_nrs if hasattr(reg, 'trgl_nrs') else None) for reg in regions],"n regions", len(regions), "check edges", len(check_edges_for_regions), "checktree", check_tree)

        # def determine_offset_region_for_pts(
        #     pts: _np_array   
        # ):
            
        #     return determine_offset_region_for_pts_inner(pts=pts, check_tree=check_tree)
        # offset_reg_id_comb_to_check[region_comb] = determine_offset_region_for_pts
        # if type(check_tree) != dict:
        #     raise ValueError("NOT A DICT")
        offset_reg_id_comb_to_check[region_comb] = lambda pts, check_tree=check_tree: determine_offset_region_for_pts_inner(pts=pts, check_tree=check_tree)

    return offset_reg_id_comb_to_check 
#

def increase_raster_to_offset_region_precision(
        raster_cell_to_regions,
        offset_reg_id_comb_to_check,
        offset_x_bins:list,
        offset_y_bins:list,
        id_to_offset_regions:dict=None,
        plot_offset_raster:bool=False,
):
    """
    Some can not be simplified: that is if one of the 4 corners of raster cell is a region vertex. hm not sure.
    """
    precise_raster = dict()

    offset_x_bins = offset_x_bins
    offset_y_bins = offset_y_bins
    lenchanges = set()
    all_precise_ids = set()
    for (ix, iy), regions_at_raster_cell in raster_cell_to_regions.items():
        if len(regions_at_raster_cell)>1:
            xmin, xmax = offset_x_bins[ix-1]
            ymin, ymax = offset_y_bins[iy-1]
            reg_ids = tuple(sorted([reg.id for reg in regions_at_raster_cell]))
            check = offset_reg_id_comb_to_check[reg_ids]
            pts = _np_array([(xmin,ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)])
            check_res = check(pts=pts)
            precise_ids = set(check_res)
            all_precise_ids.add(tuple(sorted(precise_ids)))
            preceise_regions_at_raster = [reg for reg in regions_at_raster_cell if reg.id in precise_ids]
            lenchanges.add((len(reg_ids), len(preceise_regions_at_raster)))
            if len(preceise_regions_at_raster) == 0:
                OffsetRegion.plot_many([reg for reg in regions_at_raster_cell]+[id_to_offset_regions[id] for id in precise_ids], plot_vertices=False)
                raise ValueError("regions_at_raster_cell", len(regions_at_raster_cell), [reg.id for reg in regions_at_raster_cell], "checkres", check_res, "precise_ids", precise_ids)

            precise_raster[(ix, iy)] = preceise_regions_at_raster
        else:
            precise_raster[(ix, iy)] = regions_at_raster_cell

    if not plot_offset_raster is None:
        create_raster_plot(regions=list(id_to_offset_regions.values()), raster_cell_to_regions=precise_raster, offset_x_bins=offset_x_bins, offset_y_bins=offset_y_bins)
    
    return precise_raster
#

def transform_raster_to_remaining_triangle(
        id_to_offset_regions:dict, 
        raster_cell_to_regions:dict,
        offset_x_bins:list,
        offset_y_bins:list,
        translate_reg_nr_to_reg_id:dict,
        plot_offset_raster:dict=None
    ):
    
    raster_cell_id_to_bounds = dict()
    unique_reg_id_combs_in_raster_cells = dict()
    
    for sign_x in [-1, 1]:
        for sign_y in [-1, 1]:
            # sorting 1 if abs(x)>abs(y) else -1
            for sorting in [1, -1]:
                triangle_nr = (
                    (
                        1 if sorting>0 else 2
                    ) if sign_y>0 else (
                        8 if sorting>0 else 7
                    )
                ) if sign_x>0 else (
                    (
                        4 if sorting>0 else 3
                    ) if sign_y>0 else (
                        5 if sorting>0 else 6 
                    )
                )

                for ix, (x_low, x_up) in zip(range(1,len(offset_x_bins)+1), offset_x_bins):
                    for iy, (y_low, y_up) in zip(range(1,len(offset_y_bins)+1), offset_y_bins):
                        if iy > ix:
                            break 
                        regions_at_raster_cell = raster_cell_to_regions[(ix, iy)]
                        # raster_cell_to_regions[(ix*sign_x, iy*sign_y) if sorting>0 else (iy*sign_y, ix*sign_x)] = list(set([translate_reg_nr_to_reg_id[region_nr] for region_nr in region_nrs]))
                        i_min, i_max = (iy, ix) if sorting > 0 else (ix, iy)
                        x, y = (i_max*sign_x, i_min*sign_y)
                        raster_cell_id_to_bounds[(x,y)] = ((x_low, x_up), (y_low, y_up))
                        region_nrs = [reg.nr for reg in regions_at_raster_cell] # if abs(reg.trgl_nr - triangle_nr)%7 <= 1 else reg.nr+1 
                        # try:
                        # region_ids = []
                        # for region_nr in region_nrs:
                        #     new_nr = region_nr//10*10+triangle_nr
                        #     new_reg = id_to_offset_regions[translate_reg_nr_to_reg_id[new_nr]]
                        #     if abs(new_reg.trgl_nr  - triangle_nr)%7 <= 1:
                        #         pass
                        #         if hasattr(new_reg, 'trgl_nrs'):
                        #             pass
                        #             # print("ATTT", triangle_nr,  new_reg.trgl_nr, new_reg.trgl_nrs)
                        #     else:
                        #         old_new_nr = new_nr
                        #         old_reg = new_reg
                        #         new_nr += 1 if triangle_nr < 8 else -7
                        #         new_reg = id_to_offset_regions[translate_reg_nr_to_reg_id[new_nr]]
                        #         # print(
                        #         #     "TN", triangle_nr, "new",new_nr%10, "old",old_new_nr%10,  
                        #         #     "\nnew:", (new_reg.trgl_nr, new_reg.trgl_nrs if hasattr(new_reg,'trgl_nrs') else ''),
                        #         #     "\nold:", (old_reg.trgl_nr, old_reg.trgl_nrs if hasattr(old_reg,'trgl_nrs') else ''))
                        #     region_ids.append(translate_reg_nr_to_reg_id[new_nr])
                        # region_ids = tuple(region_ids)

                        region_ids = tuple(sorted(set([translate_reg_nr_to_reg_id[region_nr//10*10+triangle_nr] for region_nr in region_nrs])))
                        # except:
                        #     print("translate_reg_nr_to_reg_id", translate_reg_nr_to_reg_id)
                        #     raise ValueError("region_nrs", region_nrs, "A",[region_nr//10*10+triangle_nr for region_nr in region_nrs])
                        if not region_ids in unique_reg_id_combs_in_raster_cells:
                            unique_reg_id_combs_in_raster_cells[region_ids] = []
                        unique_reg_id_combs_in_raster_cells[region_ids].append((x, y))

                        raster_cell_to_regions[(x,y)] = [id_to_offset_regions[id] for id in region_ids]

    if not plot_offset_raster is None:
        create_raster_plot(
            regions=list(id_to_offset_regions.values()),
            raster_cell_to_regions=raster_cell_to_regions,
            offset_x_bins=offset_x_bins,
            offset_y_bins=offset_y_bins,
            lims=[-.55,.55], add_raster_labels=True)

    offset_x_vals = get_vals_from_bins(offset_x_bins) 
    offset_y_vals = get_vals_from_bins(offset_y_bins) 
    offset_all_x_vals = [-x for x in reversed(offset_x_vals[1:])] + offset_x_vals
    offset_all_y_vals = [-y for y in reversed(offset_y_vals[1:])] + offset_y_vals

    
    return raster_cell_to_regions, offset_all_x_vals, offset_all_y_vals, unique_reg_id_combs_in_raster_cells
#


def create_region_comb_nr_to_check_lookup(
        raster_cell_to_regions,
        offset_reg_id_comb_to_check,
):
    """
    
    """
    
    # region_comb_to_nr = dict()
    # if not region_ids in region_comb_to_nr:
    #     region_comb_to_nr[region_ids] = len(region_comb_to_nr)
    raster_cell_to_region_comb_nr = dict()
    offset_region_comb_nr_to_check = dict()
    region_comb_to_nr = dict()
    for key, regions in raster_cell_to_regions.items():
        region_ids = tuple([reg.id for reg in regions])
        if not region_ids in region_comb_to_nr:
            region_comb_to_nr[region_ids] = len(region_comb_to_nr)
            offset_region_comb_nr_to_check[region_comb_to_nr[region_ids]] = offset_reg_id_comb_to_check[region_ids]
        #
        raster_cell_to_region_comb_nr[key] = region_comb_to_nr[region_ids]
    return raster_cell_to_region_comb_nr, offset_region_comb_nr_to_check
#

@time_func_perf
def prepare_offset_regions(
        grid_spacing:float,
        r:float,
        include_boundary:bool=False,
        plot_offset_checks:dict=None,
        plot_offset_regions:dict=None,
        plot_offset_raster:dict=None,
        silent:bool=True,
):
    cells_contained_in_all_disks, cells_contained_in_all_trgl_disks, cells_maybe_overlapping_a_disk, cells_maybe_overlapping_a_trgl_disk = get_cells_relevant_for_disk_by_type(
        grid_spacing=grid_spacing, r=r, include_boundary=False
        )

    trgl_regions = create_triangle_1_region()
    check_dict, cells_always_overlapped = create_check_dict(cells_maybe_overlapping_a_trgl_disk, trgl_regions)
    apply_checks_to_create_regions(check_dict=check_dict, trgl_regions=trgl_regions, r=r, plot_offset_checks=plot_offset_checks)
    
    cleanup_region_check_results(
        trgl_regions=trgl_regions, 
        cells_contained_in_all_trgl_disks=cells_contained_in_all_trgl_disks, 
        cells_always_overlapped=cells_always_overlapped,
        all_cells=cells_maybe_overlapping_a_trgl_disk,
        plot_offset_regions=plot_offset_regions,
        r=r,
        grid_spacing=grid_spacing,
    )
    
    # OffsetRegion.plot_many(regions=list(trgl_regions.values()))
    id_to_offset_regions, translate_reg_nr_to_reg_id, contain_region_mult = transform_region_to_remaining_triangles(trgl_regions=trgl_regions)
    
    (trgl_raster_cell_to_region,
     offset_x_bins,
     offset_y_bins,
     unique_reg_id_combs_in_trgl_raster_cells,
     ) = sort_trgl_region_into_raster(trgl_regions=id_to_offset_regions, plot_offset_raster=plot_offset_raster)
    
    offset_reg_id_comb_to_check = prepare_raster_to_offset_region_checks(
        id_to_offset_regions=id_to_offset_regions,
        unique_reg_id_combs_in_raster_cells=unique_reg_id_combs_in_trgl_raster_cells,
        r=r,
        include_boundary=True,
        )

    trgl_precise_raster_cell = increase_raster_to_offset_region_precision(
        raster_cell_to_regions=trgl_raster_cell_to_region,
        offset_reg_id_comb_to_check=offset_reg_id_comb_to_check,
        offset_x_bins=offset_x_bins,
        offset_y_bins=offset_y_bins,
        id_to_offset_regions=id_to_offset_regions,
        plot_offset_raster=plot_offset_raster,
        )
    

    (raster_cell_to_regions,
     offset_all_x_vals,
     offset_all_y_vals,
     unique_reg_id_combs_in_raster_cells
     ) = transform_raster_to_remaining_triangle(
         id_to_offset_regions=id_to_offset_regions,
         raster_cell_to_regions=trgl_precise_raster_cell,
         translate_reg_nr_to_reg_id=translate_reg_nr_to_reg_id,
         offset_x_bins=offset_x_bins,
         offset_y_bins=offset_y_bins,
         plot_offset_raster=plot_offset_raster)
    
    offset_reg_id_comb_to_check = prepare_raster_to_offset_region_checks( 
        id_to_offset_regions=id_to_offset_regions,
        unique_reg_id_combs_in_raster_cells=unique_reg_id_combs_in_raster_cells,
        r=r,
        include_boundary=include_boundary,
        )
    
    raster_cell_to_region_comb_nr, offset_region_comb_nr_to_check = create_region_comb_nr_to_check_lookup(
        raster_cell_to_regions=raster_cell_to_regions,
        offset_reg_id_comb_to_check=offset_reg_id_comb_to_check
    )
    
    return raster_cell_to_region_comb_nr, offset_region_comb_nr_to_check, offset_all_x_vals, offset_all_y_vals, id_to_offset_regions, contain_region_mult

